#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
GOP Keras layer definition 


Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""
from keras import backend as K
from keras.engine.topology import Layer
from keras import constraints, regularizers
from ..utility import misc, gop_operators
import tensorflow as tf

class GOP(Layer):
    """Generalized Operational Perceptron
    
    Args:
        units (int): Number of neurons
        op_set (tuple): Operator set with format (nodal, pool, activation)
        regularizer (float): Weight decay coefficient, default None
        constraint (float): Max-norm constraint, default None
        use_bias (bool): Use bias if True
        trainable (bool): Allow weight updates if True
    
        
    """
    def __init__(self,
                 units,
                 op_set,
                 regularizer=None,
                 constraint=None, 
                 use_bias = True, 
                 trainable = True,
                 **kwargs):
        
        self.units = units
        self.nodal = gop_operators.get_nodal_operator(op_set[0])
        self.pool = gop_operators.get_pool_operator(op_set[1])
        self.activation = gop_operators.get_activation_operator(op_set[2])
        self.regularizer = regularizers.l2(regularizer) if regularizer is not None else None
        self.constraint = constraints.max_norm(constraint, axis=1) if constraint is not None else None
        self.use_bias = use_bias
        self.trainable = trainable
        
        super(GOP, self).__init__(**kwargs)

    def build(self, input_shape):
        
        self.W = self.add_weight(name='W', 
                                      shape=(1, input_shape[-1], self.units),
                                      initializer=misc.get_he_init(input_shape[-1], self.units),
                                      regularizer=self.regularizer,
                                      constraint=self.constraint,
                                      trainable=self.trainable)
        if self.use_bias:
            self.bias = self.add_weight(name='bias', 
                                          shape=(self.units,),
                                          initializer='zeros',
                                          trainable=self.trainable)
            
        super(GOP, self).build(input_shape)  

    def call(self, x):
        x = tf.expand_dims(x,axis=-1)

        x = self.nodal(x, self.W)
        x = self.pool(x)
        
        if self.use_bias:
            x=K.bias_add(x,self.bias)
        
        x=self.activation(x)
        return x

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.units)
    


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Utility for training GOP-based models

Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""

from . import gop_operators, misc
import os, glob, dill, copy, subprocess, time, numpy as np
dill.settings['recurse'] = True

try:
    import cPickle as pickle
except ImportError:
    import pickle
    
CUDA_FLAG = 'CUDA_VISIBLE_DEVICES'
    

def he_init(fan_in, fan_out, shape):
    """Initialize layer weights using He init method
    
    Args:
        fan_in (int): Number of input neurons
        fan_out (int): Number of output neurons
        shape (list): Shape of layer weight
    
    Returns:
        Numpy array of given shape
        
    Reference:
        https://arxiv.org/abs/1502.01852
        
    """
    
    scale=1.0/max(1.0,float(fan_in+fan_out)/2)
    limit = np.sqrt(3. * scale)
    
    return np.random.normal(-limit, limit, shape).astype('float32')
    
    
def get_random_gop_weight(input_dim, 
                      output_dim, 
                      use_bias=True):
    """Initialize weights for a GOP block
    
    Args:
        input_dim (int): Input dimension
        output_dim (int): Output dimension
        use_bias (bool): True if using bias, False otherwise
    
    Returns:
        List of numpy arrays
        
    """
    
    W = np.random.uniform(-2*np.pi, 2*np.pi, (1, input_dim, output_dim)).astype('float32') 
    
    if use_bias:
        bias = np.zeros((output_dim,), dtype=np.float32)
        weight = [W, bias]
    else:
        weight = [W,]
        
    return weight

def get_random_block_weights(prev_block):
    """Create GOP block with random weights with shape similar to a given block
    
    Args:
        prev_block (list): Given block weights as a reference
    
    Returns:
        List of numpy arrays
    """
    prev_gop, prev_bn, prev_output = prev_block
    
    gop_input = prev_gop[0].shape[1]
    gop_output = prev_gop[0].shape[2]
    gop = get_random_gop_weight(gop_input, gop_output, True if len(prev_gop)==2 else False)
    
    bn = [np.ones((gop_output,), dtype=np.float32), 
          np.zeros((gop_output,), dtype=np.float32), 
          np.zeros((gop_output,), dtype=np.float32),
          np.ones((gop_output,), dtype=np.float32)]
    
    hidden_dim, output_dim = prev_output[0].shape
    
    o_shape = [gop_output, output_dim]
    
    output = [np.concatenate((prev_output[0], he_init(o_shape[0], o_shape[1], o_shape)), axis=0)]
    
    if len(prev_output)==2:
        output.append(prev_output[1])
        
    return gop, bn, output

def calculate_memory_block(params,
                           train_states,
                           train_func,
                           train_data):
    """Calculate memory block for POPmemH and POPmemO model
    
    Args:
        params (dict): Model parameters given as dictionary
        train_states (dict): Current topology configuration
        train_func (function): Function to generate train generator and train steps
        train_data: Input to train_func
        
    Returns:
        List of numpy arrays of memory block weights
    """
    if CUDA_FLAG in os.environ.keys():
        cuda_status = os.environ[CUDA_FLAG]
    else:
        cuda_status = ''
        
    if params['finetune_computation'][0] == 'cpu':
        os.environ[CUDA_FLAG] = ''
    else:
        os.environ[CUDA_FLAG] = misc.get_gpu_str(params['search_computation'][1])
    
    try:
        if params['memory_type'] == 'PCA':
            pre_bn_weight, projection, post_bn_weight = PCA(params, train_states, train_func, train_data)
    
        elif params['memory_type'] == 'LDA':
            pre_bn_weight, projection, post_bn_weight = LDA(params, train_states, train_func, train_data)
        else:
            raise Exception('Only supported 2 memory schemes: "PCA" and "LDA", given %s' % params['memory_type'])
            
    except:
        os.environ[CUDA_FLAG] = cuda_status
        raise Exception('Failed to calculate memory block')
    
    os.environ[CUDA_FLAG] = cuda_status
    
    return pre_bn_weight, projection, post_bn_weight

def PCA(params, train_states, train_func, train_data):
    """PCA as a memory block
    
    Args:
        params (dict): Model parameters given as dictionary
        train_states (dict): Current topology configuration
        train_func (function): Function to generate train generator and train steps
        train_data: Input to train_func
        
    Returns:
        List of numpy arrays of memory block weights
    """

    from keras.models import Model
    from keras import backend as K
    
    layer_iter = train_states['layer_iter']
    input_dim = train_states['topology'][0]
    gen, steps = train_func(train_data)
    epsilon = K.epsilon()
    op_sets = misc.map_operator_from_index(train_states['op_set_indices'], params['nodal_set'], params['pool_set'], params['activation_set'])
    
    if layer_iter > 0:
            
        model = network_builder(train_states['topology'],
                                op_sets,
                                params['input_dropout'],
                                params['dropout'],
                                params['weight_regularizer'],
                                params['weight_constraint'],
                                params['output_activation'],
                                params['use_bias'])
        
        model.compile(params['optimizer'], params['loss'], params['metrics'])
        
        weights = train_states['weights']
        for layer_name in weights.keys():
            if layer_name != 'output':
                model.get_layer(layer_name).set_weights(weights[layer_name])
        
        hidden_layer_name = 'concat_' + str(layer_iter-1)
        hidden_model = Model(inputs=model.input, outputs=model.get_layer(hidden_layer_name).output)
        
    if params['direct_computation']:
        x = np.zeros((1, input_dim), dtype=np.float32)
        
        for _ in range(steps):
            x_, _ = next(gen)
            x = np.concatenate((x, x_), axis=0)
        
        x = x[1:]
        if layer_iter > 0:
            x = hidden_model.predict(x)
            
        x_mean = np.mean(x, axis=0, keepdims=True)
        x_var = np.var(x, axis=0, keepdims=True)
        
        
        x = (x - x_mean)/(np.sqrt(x_var + epsilon))
        
        cov = np.dot(x.T, x) + params['memory_regularizer']*np.eye(x.shape[1], x.shape[1]).astype('float32')
        
    else:
        x_mean = 0.0
        N = 0
        for _ in range(steps):
            x_, _ = next(gen)
            if layer_iter > 0:
                x_ = hidden_model.predict(x_)
            x_mean += np.sum(x_, axis=0, keepdims=True) / 1e5
            N += x_.shape[0]
        
        x_mean = x_mean*1e5 / float(N)
        
        x_var = 0.0
        for _ in range(steps):
            x_, _ = next(gen)
            if layer_iter > 0:
                x_ = hidden_model.predict(x_)
            x_var += np.sum((x_ - x_mean)**2, axis=0, keepdims=True) / float(N)
            
        
        cov = 0.0
        for _ in range(steps):
            x_, _ = next(gen)
            if layer_iter > 0:
                x_ = hidden_model.predict(x_)
            x_ = (x_ - x_mean)/(np.sqrt(x_var + epsilon))
            cov += np.dot(x_.T, x_)
        
        cov += params['memory_regularizer']*np.eye(cov.shape[0])
        
    U,S,V = np.linalg.svd(cov)
    energy = np.cumsum(S)/ np.sum(S)
    idx = np.where(energy > params['min_energy_percentage'])[0][0]
    
    if idx < S.size-1:
        idx = idx +1
    
    projection = [U[:,:idx],] 
    pre_bn_weight = [np.ones((x_mean.size,), dtype=np.float32), 
                     np.zeros((x_mean.size,), dtype=np.float32), 
                     x_mean.flatten(), 
                     x_var.flatten()]
    
    bn_weight = [np.ones((idx,), dtype=np.float32), 
                 np.zeros((idx,), dtype=np.float32), 
                 np.zeros((idx,), dtype=np.float32), 
                 np.ones((idx,), dtype=np.float32)]
    
    if params['use_bias']:
        projection.append(np.zeros((idx,), dtype='float32'))
        

    K.clear_session()
    del Model
    del K
            
    return pre_bn_weight, projection, bn_weight 

def LDA(params, train_states, train_func, train_data):
    """LDA as a memory block
    
    Args:
        params (dict): Model parameters given as dictionary
        train_states (dict): Current topology configuration
        train_func (function): Function to generate train generator and train steps
        train_data: Input to train_func
        
    Returns:
        List of numpy arrays of memory block weights
    """

    from keras.models import Model
    from keras import backend as K
    
    layer_iter = train_states['layer_iter']
    input_dim = params['input_dim']
    output_dim = params['output_dim']
    gen, steps = train_func(train_data)
    epsilon = K.epsilon()
    op_sets = misc.map_operator_from_index(train_states['op_set_indices'], params['nodal_set'], params['pool_set'], params['activation_set'])
    
    if layer_iter > 0:
        
        model = network_builder(train_states['topology'],
                                op_sets,
                                params['input_dropout'],
                                params['dropout'],
                                params['weight_regularizer'],
                                params['weight_constraint'],
                                params['output_activation'],
                                params['use_bias'])
        
        model.compile(params['optimizer'], params['loss'], params['metrics'])
        
        weights = train_states['weights']
        for layer_name in weights.keys():
            if layer_name != 'output':
                model.get_layer(layer_name).set_weights(weights[layer_name])
        
        hidden_layer_name = 'concat_' + str(layer_iter-1)
        hidden_model = Model(inputs=model.input, outputs=model.get_layer(hidden_layer_name).output)
        
    if params['direct_computation']:
        x = np.zeros((1, input_dim), dtype=np.float32)
        y = np.zeros((1, output_dim), dtype=np.float32) 
        
        for _ in range(steps):
            x_, y_ = next(gen)
            x = np.concatenate((x, x_), axis=0)
            y = np.concatenate((y, y_), axis=0)
            
        x = x[1:]
        y = np.argmax(y[1:], axis=-1)
        if layer_iter > 0:
            x = hidden_model.predict(x)
            
        x_mean = np.mean(x, axis=0, keepdims=True)
        x_var = np.var(x, axis=0, keepdims=True)
        
        
        x = (x - x_mean)/(np.sqrt(x_var + epsilon))
        
        Sb = np.zeros((x.shape[-1], x.shape[-1]), dtype=np.float32)
        Sw = np.zeros((x.shape[-1], x.shape[-1]), dtype=np.float32)
        
        for c in range(output_dim):
            class_indices = np.where(y == c)[0]
            class_mean = np.mean(x[class_indices], axis=0, keepdims=True)
            Sb += np.dot(class_mean.T, class_mean)*len(class_indices)
            
            dif = x[class_indices] - class_mean
            Sw += np.dot(dif.T, dif)
        
        Sw += params['memory_regularizer'] * np.eye(Sw.shape[0])
        
    else:
        
        x_mean = 0.0
        class_pop = [0.0,] * output_dim
        class_means = [0.0,] * output_dim
        
        for _ in range(steps):
            x_, y_ = next(gen)
            y_ = np.argmax(y_, axis=-1)
            if layer_iter > 0:
                x_ = hidden_model.predict(x_)
            x_mean += np.sum(x_, axis=0, keepdims=True) / 1e5
            
            for c in range(output_dim):
                class_indices = np.where(y_ == c)[0]
                if len(class_indices) > 0:
                    class_means[c] += np.sum(x_[class_indices], axis=0, keepdims=True) / 1e5
                    class_pop[c] += len(class_indices)
            
        Sb = np.zeros((x_mean.size, x_mean.size), dtype=np.float32)
        
        for c in range(output_dim):
            class_means[c] = class_means[c]*1e5 / float(class_pop[c])
            Sb += np.dot(class_means[c].T, class_means[c])*class_pop[c]
            
        x_mean = x_mean*1e5 / float(np.sum(class_pop))
        
        x_var = 0.0
        for _ in range(steps):
            x_, _ = next(gen)
            if layer_iter > 0:
                x_ = hidden_model.predict(x_)
            x_var += np.sum((x_ - x_mean)**2, axis=0, keepdims=True) / float(np.sum(class_pop))
            
        
        Sw = np.zeros((x_mean.size, x_mean.size), dtype=np.float32)
        
        for _ in range(steps):
            x_, y_ = next(gen)
            y_ = np.argmax(y_, axis=-1)
            if layer_iter > 0:
                x_ = hidden_model.predict(x_)
            
            x_ = (x_ - x_mean)/(np.sqrt(x_var + epsilon))
            
            for c in range(output_dim):
                class_indices = np.where(y_ == c)[0]
                if len(class_indices) > 0:
                    dif = x_[class_indices] - class_means[c]
                    Sw += np.dot(dif.T, dif)
        
        Sw += params['memory_regularizer'] * np.eye(Sw.shape[0])
        
    U,S,V = np.linalg.svd(np.dot(np.linalg.inv(Sw), Sb))
    
    projection = [U[:,:output_dim -1],] 
    pre_bn_weight = [np.ones((x_mean.size,), dtype=np.float32), 
                     np.zeros((x_mean.size,), dtype=np.float32), 
                     x_mean.flatten(), 
                     x_var.flatten()]
    
    bn_weight = [np.ones((output_dim-1,), dtype=np.float32), 
                 np.zeros((output_dim-1,), dtype=np.float32), 
                 np.zeros((output_dim-1,), dtype=np.float32), 
                 np.ones((output_dim-1,), dtype=np.float32)]
    
    if params['use_bias']:
        projection.append(np.zeros((output_dim-1,), dtype='float32'))
        

    K.clear_session()
    del Model
    del K
        
    return pre_bn_weight, projection, bn_weight

def search_cluster(params,
                    train_states,
                    train_func,
                    train_data,
                    val_func,
                    val_data,
                    test_func,
                    test_data):
    """Perform operator set search procedure on cluster
    
    Args:
        params (dict): Model parameters given as dictionary
        train_states (dict): Current topology configuration
        train_func (function): Function to generate train generator and train steps
        train_data: Input to train_func
        val_func (function): Function to generate validation generator and steps
        val_data: Input to val_func
        test_func (function): Function to generate test generator and steps
        test_data: Input to test_func
        
    Returns:
        performance (dict): Performance of best operator set at best model setting
        weights (list): Weights of new block
        op_set (tuple): Best performing operator set
        history (dict): full history (loss, metrics) 
    """
    
    misc.dump_data(params,
                   train_states,
                   train_func,
                   train_data,
                   val_func,
                   val_data,
                   test_func,
                   test_data)
    
    runnable = 'gop_search.py'
    
    path = os.path.join(params['tmp_dir'], params['model_name'])
    batch_job_filename = os.path.join(path, 'batchjob.sh')
    
    # write batch_job files if not exist
    if not os.path.exists(batch_job_filename):
        batchjob_parameters = params['batchjob_parameters']
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), runnable)
        cmd = 'gop_search_path=%s %s %s -N %d -n "$SLURM_ARRAY_TASK_ID" \n' % (path, batchjob_parameters['python_cmd'], filename, batchjob_parameters['no_machine'])
        contents = misc.setup_batchjob_file(batchjob_parameters, cmd, path)
        
        with open(batch_job_filename,'w') as fid:
            for row in contents:
                fid.write(row)
        
    # submit batch job
    cmd = 'sbatch --parsable %s' % batch_job_filename
    try:
        job_id = subprocess.check_output(cmd, shell=True)
        job_id = job_id.decode('utf-8')
    except:
        cmd = 'sbatch %s' % batch_job_filename
        try:
            output = subprocess.check_output(cmd, shell=True)
            output = output.decode('utf-8')
            job_id = output.split(' ')[-1]
        except:
            raise Exception('Failed to submit batch job')
            
    
    # loop & wait for result
    while len(glob.glob(os.path.join(path, 'machine_*_finish.txt'))) != params['batchjob_parameters']['no_machine']:
        job_status = subprocess.check_output('squeue -j %s' % job_id, shell=True)
        job_status = job_status.decode('utf-8')
        if len(job_status.split('\n'))==2:
            raise Exception('Search process using cluster terminated but no returned result exist!\n' +\
                        'Errors potentially happen during search process')
        time.sleep(5)
    
    if params['direction'] == 'higher':
        measure = np.NINF
        sign = 1.0
    else:
        measure = np.inf
        sign = -1.0
        
    machine_index = -1
    
    with open(os.path.join(path, 'machine_result.txt'), 'r') as fid:
        content = fid.read().split('\n')[:-1]

    for row in content:
        idx, p = row.split(',')
        if sign*(float(p) - measure) > 0:
            measure = float(p)
            machine_index = int(idx)
    
    with open(os.path.join(path, 'machine_%d.pickle' % machine_index), 'rb') as fid:
        result = pickle.load(fid)
        
    performance = result['performance']
    weights = result['weights']
    op_set_idx = result['op_set_idx']
    history = result['history']
    
    # clean intermediate results
    batchjob_name = params['batchjob_parameters']['name']
    removed_files =  glob.glob(os.path.join(path, 'machine_*')) +\
                     glob.glob(os.path.join(path, batchjob_name, '*.e')) + \
                     glob.glob(os.path.join(path, batchjob_name, '*.o'))
                     
    
    misc.remove_files(removed_files)
        
    return performance, weights, op_set_idx, history

def search_cpu(params,
               train_states,
               train_func,
               train_data,
               val_func,
               val_data,
               test_func,
               test_data):
    """Perform operator set search procedure on local machine using CPU
    
    Args:
        params (dict): Model parameters given as dictionary
        train_states (dict): Current topology configuration
        train_func (function): Function to generate train generator and train steps
        train_data: Input to train_func
        val_func (function): Function to generate validation generator and steps
        val_data: Input to val_func
        test_func (function): Function to generate test generator and steps
        test_data: Input to test_func
        
    Returns:
        performance (dict): Performance of best operator set at best model setting
        weights (list): Weights of new block
        op_set (tuple): Best performing operator set
        history (dict): full history (loss, metrics) 
    """
    
    misc.dump_data(params,
                  train_states,
                  train_func,
                  train_data,
                  val_func,
                  val_data,
                  test_func,
                  test_data)
    
    runnable = 'gop_search.py'
    
    path = os.path.join(params['tmp_dir'], params['model_name'])
    filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), runnable)
    cmd = 'CUDA_VISIBLE_DEVICES="" gop_search_path=%s python %s -N %d -n %d' %(path, filename, 1, 0)
    p = subprocess.Popen(cmd, shell=True)
    
    p.wait()
    
    if not os.path.exists(os.path.join(path, 'machine_0_finish.txt')):
        raise Exception('Search process using CPU terminated but no returned result exist!\n' +\
                        'Errors potentially happen during search process')
    
    with open(os.path.join(path, 'machine_0.pickle'),'rb') as fid:
        result = pickle.load(fid)
        
    removed_files = glob.glob(os.path.join(path, 'machine*'))
    misc.remove_files(removed_files)
    
    return result['performance'], result['weights'], result['op_set_idx'], result['history']

def search_gpu(params,
               train_states,
               train_func,
               train_data,
               val_func,
               val_data,
               test_func,
               test_data):
    """Perform operator set search procedure on local machine using GPU
    
    Args:
        params (dict): Model parameters given as dictionary
        train_states (dict): Current topology configuration
        train_func (function): Function to generate train generator and train steps
        train_data: Input to train_func
        val_func (function): Function to generate validation generator and steps
        val_data: Input to val_func
        test_func (function): Function to generate test generator and steps
        test_data: Input to test_func
        
    Returns:
        performance (dict): Performance of best operator set at best model setting
        weights (list): Weights of new block
        op_set (tuple): Best performing operator set
        history (dict): full history (loss, metrics) 
    """
    misc.dump_data(params,
               train_states,
               train_func,
               train_data,
               val_func,
               val_data,
               test_func,
               test_data)
    
    path = os.path.join(params['tmp_dir'], params['model_name'])
    no_op_set = params['no_op_set']
    performance, weights, op_set_idx, history = search_gpu_(path,
                                                       params,
                                                       0,
                                                       no_op_set,
                                                       0)
    
    
    return performance, weights, op_set_idx, history
    
    
def search_gpu_(path,
                params,
                start_idx,
                stop_idx,
                machine_no):
    """Auxiliary function to distribute search jobs on different GPUs
    
    Args:
        path (str): Path to directory holding temporary data
        params (dict): Model parameters
        start_idx (int): Start index in the list of operator sets
        stop_idx (int): Stop index in the list of operator sets
        machine_no (int): Index of machine that is given start_idx, stop_idx
        
    Returns:
        performance (dict): Performance of best operator set from start_idx to stop_idx
        weights (list): Weights of new block
        op_set (tuple): Best performing operator set
        history (dict): full history (loss, metrics) 
        
    """

    start_indices, stop_indices = misc.partition_indices(start_idx, stop_idx, len(params['search_computation'][1]))
    
    runnable = 'gop_search_gpu.py'
    
    P = []
    for gpu, b, e in zip(params['search_computation'][1], start_indices, stop_indices):
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), runnable)
        cmd = 'gop_search_path=%s gop_machine_no=%d CUDA_VISIBLE_DEVICES=%d python %s -b %d -e %d' %(path,machine_no, gpu, filename, b, e)
        print(cmd)
        P.append(subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE))
        
    for p in P:
        p.wait()
        
    if len(glob.glob(os.path.join(path,'machine_%d_*_finish.txt' % machine_no))) != len(start_indices):
        raise Exception('Search process using GPU terminated but no returned result exist\n' + \
                        'Errors potentially happen during search process')
    
    with open(os.path.join(path, 'machine_%d_result.txt' % machine_no), 'r') as fid:
        content = fid.read().split('\n')[:-1]
    
    if params['direction'] == 'higher':
        measure = np.NINF
        sign = 1.0
    else:
        measure = np.inf
        sign = -1.0
        
    gpu = -1
    
    for row in content:
        idx, p = row.split(',')
        if sign*(float(p)-measure) > 0:
            measure = float(p)
            gpu = int(idx)
    
    with open(os.path.join(path, 'machine_%d_%d.pickle' % (machine_no, gpu)), 'rb') as fid:
        irsgpu_result = pickle.load(fid)
        
    performance = irsgpu_result['performance']
    weights = irsgpu_result['weights']
    op_set_idx = irsgpu_result['op_set_idx']
    history = irsgpu_result['history']
    
    removed_files = glob.glob(os.path.join(path, 'machine_*'))
    misc.remove_files(removed_files)
    
    return performance, weights, op_set_idx, history
    

def GIS(start_idx,
        stop_idx,
        params,
        train_states):
    """Greedy Iterative Search for POP
    
    Perform one iterative search pass for POP algorithm
    
    Args:
        start_idx (int): Start index in the list of operator sets
        stop_idx (int): Stop index in the list of operator sets
        params (dict): Model parameters
        train_states (dict): Current topology configuration
        
    Returns:
        performance (dict): Performance of best operator set from start_idx to stop_idx
        weights (list): Weights of new block
        op_set (tuple): Best performing operator set
        history (dict): full history (loss, metrics) 
        
    """

    from joblib import Parallel, delayed
        
    
    op_set_indices = range(params['no_op_set'])
    
    if start_idx is not None:
        all_op_set_indices = op_set_indices[start_idx:stop_idx]
        
    if params['search_computation'][0] == 'cpu':
        no_process = params['search_computation'][1]
    else:
        no_process = 1
    
    search_results = Parallel(n_jobs=no_process, temp_folder=os.path.join(params['tmp_dir'], params['model_name']))\
            (delayed(GIS_)(params, train_states, op_set_idx) for op_set_idx in all_op_set_indices)
    
    performance, weights, op_set_idx, history = get_optimal_op_set(search_results, params['convergence_measure'], params['direction'])
    
    return performance, weights, op_set_idx, history


    return

def GIS_(params,
         train_states,
         op_set_idx):
    """Auxiliary function of GIS to evaluate the given operator set
    
    Args:
        params (dict): Model parameters
        train_states (dict): Current topology setting
        op_set_idx (int): Operator set index for new block
    
    Returns:
        block_performance (dict): Performance when adding new block
        weights (list): Weights of new block
        op_set_idx (int): The given op_set index
        history (dict): full history (loss, metrics) 
    """
    
    path = os.path.join(params['tmp_dir'], params['model_name'])
    train_func, train_data, val_func, val_data, test_func, test_data = misc.unpickle_generator(os.path.join(path, 'data.pickle'))
    
    if os.path.exists(os.path.join(path, 'custom_loss.pickle')):
        params['loss'] = misc.unpickle_custom_loss(os.path.join(path, 'custom_loss.pickle'))
    
    if os.path.exists(os.path.join(path, 'custom_metrics.pickle')):
        params['metrics'] = misc.unpickle_custom_metrics(os.path.join(path, 'custom_metrics.pickle'))
        
    if os.path.exists(os.path.join(path, 'special_metrics.pickle')):
        params['special_metrics'] = misc.unpickle_special_metrics(os.path.join(path, 'special_metrics.pickle'))
        
    if os.path.exists(os.path.join(path, 'custom_operators.pickle')):
        params['nodal_set'], params['pool_set'], params['activation_set'] = misc.unpickle_custom_operators(os.path.join(path, 'custom_operators.pickle'))
    
    op_set_indices = copy.deepcopy(train_states['op_set_indices'])
    weights = copy.deepcopy(train_states['weights'])
    layer_iter = train_states['layer_iter']
    
    suffix = '_' + str(layer_iter) + '_0' 
    
    if train_states['search_layer'] == 'output':
        op_set_indices['output'] = op_set_idx
        op_set_indices['gop' + suffix] = train_states['hidden_op_set_idx']
    else:
        op_set_indices['gop' + suffix] = op_set_idx
        op_set_indices['output'] = train_states['output_op_set_idx']
    
    prev_hidden_dim = params['input_dim'] if layer_iter == 0 else params['max_topology'][layer_iter-1]
    block_size = params['max_topology'][layer_iter]
    
    weights['gop' + suffix] = get_random_gop_weight(prev_hidden_dim, block_size, params['use_bias'])
    weights['bn' + suffix] = [np.ones((block_size,), dtype=np.float32),
                              np.zeros((block_size,), dtype=np.float32),
                              np.zeros((block_size,), dtype=np.float32),
                              np.ones((block_size,), dtype=np.float32),]
    weights['output'] = get_random_gop_weight(params['max_topology'][layer_iter], params['output_dim'], params['use_bias'])        
    
    block_names = ['gop'+suffix, 'bn'+suffix, 'output']
    
    
    
    
    measure, history, weights = block_update(train_states['topology'], 
                                             op_set_indices, 
                                             weights, 
                                             params, 
                                             block_names, 
                                             train_func,
                                             train_data,
                                             val_func,
                                             val_data,
                                             test_func,
                                             test_data)
    
    
    block_weights = [weights[block_names[0]], weights[block_names[1]], weights[block_names[2]]]
    
    idx = np.argmax(history[params['convergence_measure']]) if params['direction'] == 'higher' else np.argmin(history[params['convergence_measure']])
    block_performance = {}
    
    for metric in history.keys():
        block_performance[metric] = history[metric][idx]
        
    
    return block_performance, block_weights, op_set_idx, history


def GISfast(start_idx,
            stop_idx,
            params, 
            train_states):
    """Fast version of Greedy Iterative Search
    
    This search procedure assumes the output layer is a linear layer
    
    Args:
        start_idx (int): Start index in the list of operator sets
        stop_idx (int): Stop index in the list of operator sets
        params (dict): Model parameters
        train_states (dict): Current topology configuration
        
    Returns:
        performance (dict): Performance of best operator set from start_idx to stop_idx
        weights (list): Weights of new block
        op_set (tuple): Best performing operator set
        history (dict): full history (loss, metrics) 
    
    """
    
    from joblib import Parallel, delayed
        
    
    all_op_set_indices = range(params['no_op_set'])
    
    if start_idx is not None:
        all_op_set_indices = all_op_set_indices[start_idx:stop_idx]
        
    if params['search_computation'][0] == 'cpu':
        no_process = params['search_computation'][1]
    else:
        no_process = 1
    
    search_results = Parallel(n_jobs=no_process, temp_folder=os.path.join(params['tmp_dir'], params['model_name']))\
            (delayed(GISfast_)(params, train_states, op_set_idx) for op_set_idx in all_op_set_indices)
    
    performance, weights, op_set_idx, history = get_optimal_op_set(search_results, params['convergence_measure'], params['direction'])
    
    return performance, weights, op_set_idx, history


def GISfast_(params,
             train_states,
             op_set_idx):
    """Auxiliary function of GISfast to evaluate the given operator set
    
    Args:
        params (dict): Model parameters
        train_states (dict): Current topology setting
        op_set_idx (int): Operator set index for new block
    
    Returns:
        block_performance (dict): Performance when adding new block
        weights (list): Weights of new block
        op_set_idx (int): The given op_set_idx
        history (dict): full history (loss, metrics) 
    """
    
    path = os.path.join(params['tmp_dir'], params['model_name'])
    train_func, train_data, val_func, val_data, test_func, test_data = misc.unpickle_generator(os.path.join(path, 'data.pickle'))
    
    if os.path.exists(os.path.join(path, 'custom_loss.pickle')):
        params['loss'] = misc.unpickle_custom_loss(os.path.join(path, 'custom_loss.pickle'))
    
    if os.path.exists(os.path.join(path, 'custom_metrics.pickle')):
        params['metrics'] = misc.unpickle_custom_metrics(os.path.join(path, 'custom_metrics.pickle'))
        
    if os.path.exists(os.path.join(path, 'special_metrics.pickle')):
        params['special_metrics'] = misc.unpickle_special_metrics(os.path.join(path, 'special_metrics.pickle'))
        
    if os.path.exists(os.path.join(path, 'custom_operators.pickle')):
        params['nodal_set'], params['pool_set'], params['activation_set'] = misc.unpickle_custom_operators(os.path.join(path, 'custom_operators.pickle'))
        
    layer_iter = train_states['layer_iter']
    block_iter = train_states['block_iter']
    block_suffix = '_' + str(layer_iter) + '_' + str(block_iter)
    block_size = params['max_topology'][layer_iter]
    
    if layer_iter == 0:
        prev_hidden_dim = train_states['topology'][0]
    else:
        prev_hidden_dim = 0
        for block in train_states['topology'][-3]:
            prev_hidden_dim += block[1]
    
    # add new gop block
    topology = copy.deepcopy(train_states['topology'])
    topology[-2].append(('gop', block_size))
    
    hidden_dim = 0
    for block in topology[-2]:
        hidden_dim += block[1]
        
    output_dim = topology[-1][1]
    
    op_set_indices = copy.deepcopy(train_states['op_set_indices'])
    op_set_indices['gop' + block_suffix] = op_set_idx
    
    
    
    weights = copy.deepcopy(train_states['weights'])
    gop_weight = [he_init(prev_hidden_dim, block_size, (1, prev_hidden_dim, block_size)),]
    
    bn_weight = [np.ones((block_size,), dtype=np.float32), 
                 np.zeros((block_size,), dtype=np.float32), 
                 np.zeros((block_size,), dtype=np.float32),
                 np.ones((block_size,), dtype=np.float32)]
    
    output_weight = [he_init(hidden_dim, output_dim, [hidden_dim, output_dim])]
    
    if params['use_bias']:
        gop_weight.append(np.zeros((block_size,), dtype=np.float32))
        output_weight.append(np.zeros((output_dim,), dtype=np.float32))
        
    weights['gop' + block_suffix] = gop_weight
    weights['bn' + block_suffix] = bn_weight
    weights['output'] = output_weight
        
    block_names = ['gop' + block_suffix,
                   'bn' + block_suffix,
                   'output']

    measure, history, weights = block_update(topology, 
                                             op_set_indices, 
                                             weights, 
                                             params, 
                                             block_names, 
                                             train_func,
                                             train_data,
                                             val_func,
                                             val_data,
                                             test_func,
                                             test_data)
    
    
    block_weights = [weights[block_names[0]], weights[block_names[1]], weights[block_names[2]]]
    
    idx = np.argmax(history[params['convergence_measure']]) if params['direction'] == 'higher' else np.argmin(history[params['convergence_measure']])
    block_performance = {}
    
    for metric in history.keys():
        block_performance[metric] = history[metric][idx]
        
    
    return block_performance, block_weights, op_set_idx, history

def IRS(start_idx,
        stop_idx,
        params, 
        train_states):
    """Iterative Randomized Search for HeMLGOP, HoMLGOP, HeMLRN, HoMLRN
    
    This function performs the evaluation of a set of operator sets using
    randomized method
    
    Args:
        start_idx (int): Start index in the list of operator sets
        stop_idx (int): Stop index in the list of operator sets
        params (dict): Model parameters
        train_states (dict): Current topology configuration
        
    Returns:
        performance (dict): Performance of best operator set from start_idx to stop_idx
        weights (list): Weights of new block
        op_set_idx (int): Index of best performing operator set
        history (dict): full history (loss, metrics) 
    
    """
    
    from joblib import Parallel, delayed
    
    op_set_indices = range(params['no_op_set'])
    
    if start_idx is not None:
        all_op_set_indices = op_set_indices[start_idx:stop_idx]
        
    if params['search_computation'][0] == 'cpu':
        no_process = params['search_computation'][1]
    else:
        no_process = 1
    
    search_results = Parallel(n_jobs=no_process, temp_folder=os.path.join(params['tmp_dir'], params['model_name']))\
            (delayed(LS)(params, train_states, op_set_idx) for op_set_idx in all_op_set_indices)
    
    performance, weights, op_set_idx, history = get_optimal_op_set(search_results, params['convergence_measure'], params['direction'])
    
    return performance, weights, op_set_idx, history


def LS(params, 
       train_states, 
       op_set_idx):
    """Auxiliary function of IRS to evaluate the given operator set
    
    Args:
        params (dict): Model parameters
        train_states (dict): Current topology setting
        op_set (int): Index of operator set of new block
    
    Returns:
        block_performance (dict): Performance when adding new block
        weights (list): Weights of new block
        op_set_idx (int): Index of the given operator set
        history (dict): full history (loss, metrics) 
    """
            
    from keras.models import Model
    from keras import backend as K
    
    path = os.path.join(params['tmp_dir'], params['model_name'])
    train_func, train_data, val_func, val_data, test_func, test_data = misc.unpickle_generator(os.path.join(path, 'data.pickle'))
    
    if os.path.exists(os.path.join(path, 'custom_loss.pickle')):
        params['loss'] = misc.unpickle_custom_loss(os.path.join(path, 'custom_loss.pickle'))
    
    if os.path.exists(os.path.join(path, 'custom_metrics.pickle')):
        params['metrics'] = misc.unpickle_custom_metrics(os.path.join(path, 'custom_metrics.pickle'))
        
    if os.path.exists(os.path.join(path, 'special_metrics.pickle')):
        params['special_metrics'] = misc.unpickle_special_metrics(os.path.join(path, 'special_metrics.pickle'))
        
    if os.path.exists(os.path.join(path, 'custom_operators.pickle')):
        params['nodal_set'], params['pool_set'], params['activation_set'] = misc.unpickle_custom_operators(os.path.join(path, 'custom_operators.pickle'))
    
    
    train_gen, train_steps = train_func(train_data)
    
    if val_func:
        val_gen, val_steps = val_func(val_data)
    else:
        val_gen, val_steps = None, None
        
    if test_func:
        test_gen, test_steps = test_func(test_data)
    else:
        test_gen, test_steps = None, None
        
    # get parameters
    layer_iter = train_states['layer_iter']
    block_iter = train_states['block_iter']
    use_bias = params['use_bias']
    block_suffix = '_' + str(layer_iter) + '_' + str(block_iter)
    block_size = params['block_size']
    weights = train_states['weights']
        
    # add new gop block
    topology = copy.deepcopy(train_states['topology'])
    topology[-2].append(('gop', block_size))
    op_set_indices = copy.deepcopy(train_states['op_set_indices'])
    op_set_indices['gop' + block_suffix] = op_set_idx
    op_sets = misc.map_operator_from_index(op_set_indices, params['nodal_set'], params['pool_set'], params['activation_set'])
    
    model = network_builder(topology,
                            op_sets,
                            params['input_dropout'],
                            params['dropout'],
                            params['weight_regularizer'],
                            params['weight_constraint'],
                            params['output_activation'],
                            use_bias)
    
    
    model.compile(params['optimizer'], loss=params['loss'], metrics=params['metrics'])
    
    # set fixed weights & random weights
    block_input_dim = 0
    if layer_iter == 0:
        block_input_dim = topology[0]
    else:
        for _, units in topology[-3]:
            block_input_dim += units
    
    last_hidden_dim = 0
    for _, units in topology[-2]:
        last_hidden_dim += units
        
    # set random weights for new block
    gop_weight = get_random_gop_weight(block_input_dim, block_size, use_bias)
    model.get_layer('gop' + block_suffix).set_weights(gop_weight)
    
    # set fixed weights
    for layer_name in train_states['weights'].keys():
        if layer_name != 'output':
            model.get_layer(layer_name).set_weights(weights[layer_name])
    
    hidden_layer_name = 'concat_' + str(layer_iter) if block_iter > 0 else 'bn' + block_suffix
    hidden_model = Model(inputs=model.input, outputs=model.get_layer(hidden_layer_name).output)
    
    bn_weight, output_weight = least_square(hidden_model, 
                                            topology[0], 
                                            last_hidden_dim, 
                                            block_size,
                                            topology[-1][-1], 
                                            train_gen, 
                                            train_steps, 
                                            params['direct_computation'], 
                                            params['least_square_regularizer'], 
                                            use_bias,
                                            K.epsilon())
    
    
    model.get_layer('bn' + block_suffix).set_weights(bn_weight)
    model.get_layer('output').set_weights(output_weight)
    
    block_performance = misc.evaluate(model, train_gen, train_steps, val_gen, val_steps, test_gen, test_steps)
    
    if params['special_metrics']:
        train_performance = misc.evaluate_special_metrics(model, params['special_metrics'], train_gen, train_steps)
        val_performance = misc.evaluate_special_metrics(model, params['special_metrics'], val_gen, val_steps) if val_gen else None
        test_performance = misc.evaluate_special_metrics(model, params['special_metrics'], test_gen, test_steps) if test_gen else None
        
        for idx, metric in enumerate(params['special_metrics']):
            block_performance['train_' + metric.__name__] = train_performance[idx]
            block_performance['val_' + metric.__name__] = val_performance[idx] if val_performance else None
            block_performance['test_' + metric.__name__] = test_performance[idx] if test_performance else None
    
    block_weights = [gop_weight, bn_weight, output_weight]
    
    K.clear_session()
    del Model
    del K
    return block_performance, block_weights, op_set_idx, block_performance 

def least_square(model, 
                 input_dim, 
                 hidden_dim, 
                 block_size,
                 output_dim, 
                 gen, 
                 steps, 
                 direct_computation, 
                 regularizer, 
                 use_bias,
                 epsilon):
    """Auxiliary function of LS to solve least-square problem
    
    Args:
        model: Keras model that produce hidden representation
        input_dim (int): Dimension of input data
        hidden_dim (int): Dimension of current hidden layer
        block_size (int): Size of new block
        output_dim (int): Dimension of target
        gen (generator): Data generator
        steps (int): Number of mini-batches
        direction_computation (bool): Compute using full-batch if True, otherwise mini-batch
        regularizer (float): Regularization of least-square problem
        use_bias (bool): append bias if True
        epsilon (float): small amount added to denominator to avoid zero division
        
    Returns:
        bn_weight (list): weights of BN layer in the block
        output_weight (list): weights of output layer
        
    """
    
    if direct_computation:
        x = np.zeros((1, input_dim), dtype=np.float32)
        y = np.zeros((1, output_dim), dtype=np.float32)
        
        for _ in range(steps):
            x_, y_ = next(gen)
            x = np.concatenate((x, x_), axis=0)
            y = np.concatenate((y, y_), axis=0)
        
        y = y[1:]
        x = model.predict(x[1:])
        x_mean = np.mean(x, axis=0, keepdims=True)
        x_var = np.var(x, axis=0, keepdims=True)
        
        # set all except last block to zero, one 
        x_mean[:,:hidden_dim-block_size] = 0.0
        x_var[:,:hidden_dim-block_size] = 1.0 - epsilon
        
        x = (x - x_mean)/(np.sqrt(x_var + epsilon))
        xTx = np.dot(x.T, x)
        xTy = np.dot(x.T, y)
        W = np.dot(np.linalg.pinv(xTx + regularizer*np.eye(hidden_dim, hidden_dim, dtype=np.float32)), xTy)
        
        
    else:
        x_mean = 0.0
        N = 0
        for _ in range(steps):
            x_, _ = next(gen)
            x_ = model.predict(x_)
            x_mean += np.sum(x_, axis=0, keepdims=True) / 1e5
            N += x_.shape[0]
        
        x_mean = x_mean*1e5 / float(N)
        
        x_var = 0.0
        for _ in range(steps):
            x_, _ = next(gen)
            x_ = model.predict(x_)
            x_var += np.sum((x_ - x_mean)**2, axis=0, keepdims=True) / float(N)
            
        # set all except last block to zero, one
        x_mean[:,:hidden_dim-block_size] = 0.0
        x_var[:,:hidden_dim-block_size] = 1.0 - epsilon
        
        xTx = 0.0
        xTy = 0.0
        for _ in range(steps):
            x_, y_ = next(gen)
            x_ = model.predict(x_)
            x_ = (x_ - x_mean)/(np.sqrt(x_var + epsilon))
            xTx += np.dot(x_.T, x_)
            xTy += np.dot(x_.T, y_)
    
        W = np.dot(np.linalg.pinv(xTx + regularizer*np.eye(hidden_dim, hidden_dim, dtype=np.float32)), xTy)
        
    
    
    if use_bias:
        output_weight = [W, np.zeros((output_dim,), dtype=np.float32)]
    else:
        output_weight = [W,]
        
    
    bn_weight = [np.ones((block_size,), dtype=np.float32), 
                 np.zeros((block_size,), dtype=np.float32), 
                 x_mean.flatten()[hidden_dim-block_size:], 
                 x_var.flatten()[hidden_dim-block_size:]]
    

    
    return bn_weight, output_weight

def block_update(topology, 
                 op_set_indices, 
                 weights, 
                 params, 
                 block_names, 
                 train_func,
                 train_data,
                 val_func,
                 val_data,
                 test_func,
                 test_data):
    """Construct the network and update the given blocks
    
    Args:
        topology (list): Data about the network topology
        op_set_indices (dict): Operator set indices of each block
        weights (dict): Weights of each block
        params (dict): Model parameters
        block_names (list): List of layers to update
        train_func (function): Function to return train generator and #mini-batches
        train_data: Input to train_func
        val_func (function): Function to return val generator and #mini-batches
        val_data: Input to val_func
        test_func (function): Function to return test generator and #mini-batches
        test_data: Input to test_func
        
    Returns:
        measure (float): Performance according to convergence_measure
        history (dict): History (loss, metrics) of all epochs
        weights (dict): Weights after update
        
    """
    
    op_sets = misc.map_operator_from_index(op_set_indices, params['nodal_set'], params['pool_set'], params['activation_set'])
    
    model = network_builder(topology,
                            op_sets,
                            params['input_dropout'],
                            params['dropout'],
                            params['weight_regularizer'],
                            params['weight_constraint'], 
                            params['output_activation'],
                            params['use_bias'])
                
    model.compile(params['optimizer'], params['loss'], params['metrics'])
        
    model = misc.set_model_weights(model, weights)
    
    
    for layer in model.layers:
        if layer.name in block_names:
            layer.trainable = True
        else:
            layer.trainable = False
    
    measure, history, new_weights = network_trainer(model,
                                                    params['direction'],
                                                    params['convergence_measure'],
                                                    params['lr_train'], 
                                                    params['epoch_train'],
                                                    params['optimizer'],
                                                    params['loss'],
                                                    params['metrics'],
                                                    params['special_metrics'],
                                                    train_func,
                                                    train_data,
                                                    val_func,
                                                    val_data,
                                                    test_func,
                                                    test_data)
                                    
    # get block weights
    model.set_weights(new_weights)
    for layer_name in weights.keys():
        weights[layer_name] = model.get_layer(layer_name).get_weights()
        
    from keras import backend as K
    K.clear_session()
    del K
    
    return measure, history, weights
    
    
    
def network_builder(topology,
                    op_sets,
                    input_dropout=None,
                    dropout=None,
                    regularizer=None,
                    constraint=None,
                    output_activation=None,
                    use_bias=True):
    """Building Keras model using given topology and parameters
    
    Args:
        topology (list): Data about network topology
        op_sets (dict): Operator sets of each block
        input_dropout (float): Dropout applied to input layer, default None
        dropout (float): Dropout applied to hidden layers, default None
        regularizer (float): Weight decay coefficient, default None
        constraint (float): Max-norm constraint of weights, default None
        output_activation (str): Optional activation applied to output layer, default None
        use_bias (bool): Allow using bias if True
        
    Returns:
        Keras model without compilation
        
    """
    
    from ..layers import gop
    from keras.layers import Dropout, Input, Dense, BatchNormalization as BN, Concatenate
    from keras.models import Model
    from keras import regularizers, constraints
    
    inputs = Input((topology[0],))
    
    if input_dropout:
        hiddens = Dropout(input_dropout, name='input_dropout')(inputs)
    else:
        hiddens = inputs
    
    no_layer = len(topology)
    
    for layer_idx in range(1,no_layer-1):
        layer_blocks = []
        layer_output_dim = 0
        no_block = len(topology[layer_idx])
        
        for block_idx in range(no_block):
            # if last block, use given op_set and block_size
            suffix = '_' + str(layer_idx-1) + '_' + str(block_idx)
            block_type, units = topology[layer_idx][block_idx]
            
            layer_output_dim += units
            
            if block_type == 'gop':
                hiddens_ = gop.GOP(units,
                                   op_sets[block_type + suffix],
                                   regularizer,
                                   constraint,
                                   use_bias,
                                   name='gop' + suffix)(hiddens)
                
                hiddens_ = BN(name = 'bn' + suffix)(hiddens_)
            
            elif block_type == 'mem':
                hiddens_ = BN(name='mem_pre_bn' + suffix)(hiddens)
                hiddens_ = Dense(units, use_bias=use_bias, name='mem'+suffix)(hiddens_)
                hiddens_ = BN(name='bn'+suffix)(hiddens_)
            
            else:
                raise 'Invalid block type ´´%s´´' % block_type
                
                
            layer_blocks.append(hiddens_)
            
        
        if len(layer_blocks) > 1:
            hiddens = Concatenate(axis=-1, name='concat_' + str(layer_idx-1))(layer_blocks)
        else:
            hiddens = layer_blocks[0]
        
        if dropout:
            hiddens = Dropout(dropout, name='dropout_' + str(layer_idx-1))(hiddens)
            
    output_type, output_units = topology[-1]
    
    if output_type == 'dense':
        outputs = Dense(output_units, 
                        kernel_regularizer=regularizers.l2(regularizer) if regularizer else None,
                        kernel_constraint=constraints.max_norm(constraint, axis=0) if constraint else None,
                        activation=output_activation, 
                        use_bias=use_bias, 
                        name='output')(hiddens)
    
    elif output_type == 'gop':
        outputs = gop.GOP(output_units, 
                          op_sets['output'], 
                          regularizer,
                          constraint,
                          use_bias,
                          name='output')(hiddens)
    
    model = Model(inputs=inputs, outputs=outputs)

    return model
        

def network_trainer(model,
                    direction,
                    convergence_measure,
                    LR, 
                    SC,
                    optimizer,
                    loss,
                    metrics,
                    special_metrics,
                    train_func,
                    train_data,
                    val_func,
                    val_data,
                    test_func,
                    test_data):
    """Train the given model with given parameters
    
    Args:
        model: Keras model instance
        direction (str): String indicates quality of convergence measure ('higher'/'lower')
        convergence_measure (str): Name of the metric to evaluate when to stop
        LR (list): List of learning rate schedule
        SC (list): Number of epochs for each learning rate value
        optimizer (str): Name of the optimizer
        loss (str/callable): Loss function
        metrics (list): List of metrics to evaluate
        special_metrics (list): List of special metrics to evaluate
        train_func (function): Function to return train generator and #mini-batches
        train_data: Input to train_func
        val_func (function): Function to return val generator and #mini-batches
        val_data: Input to val_func
        test_func (function): Function to return test generator and #mini-batches
        test_data: Input to test_func
        
    Returns:
        measure: The value of convergence_measure
        history: Entire history (loss/metrics) of given data
        optimal_weights: Network weights after training
        
    """
    
    # set up generator
    train_gen, train_steps = train_func(train_data)
    val_gen, val_steps = val_func(val_data) if val_func else (None, None)
        
    # set up parameters
    
    current_weights = model.get_weights()
    optimal_weights = model.get_weights()
    trainable_status = {}
    for layer in model.layers:
        trainable_status[layer.name] = layer.trainable
    
    if direction == 'higher':
        measure = np.NINF
        sign = 1.0
    else:
        measure = np.inf
        sign = -1.0
        
    history = {}
    
    for metric in model.metrics_names:
        history['train_' + metric] = []
        history['val_' + metric] = [] if val_gen else None
        history['test_' + metric] = [] if test_func else None
    
    if special_metrics:
        for metric in special_metrics:
            history['train_' + metric.__name__] = []
            history['val_' + metric.__name__] = [] if val_gen else None
            history['test_' + metric.__name__] = [] if test_func else None
    
    for lr, sc in zip(LR,SC):
        model.compile(optimizer, loss, metrics)
        model.set_weights(current_weights)
        for layer in model.layers:
            layer.trainable = trainable_status[layer.name]
        
        cb = misc.BestModelLogging(convergence_measure, 
                                   direction,
                                   special_metrics,
                                   train_func,
                                   train_data,
                                   val_func,
                                   val_data,
                                   test_func,
                                   test_data)
        #model.summary()
        model.fit_generator(train_gen,
                            train_steps,
                            epochs=sc,
                            verbose=0,
                            callbacks=[cb,],
                            validation_data=val_gen,
                            validation_steps=val_steps)
        
        if sign*(cb.measure - measure) > 0:
            optimal_weights = cb.model_weights
            measure = cb.measure
        
        current_weights = model.get_weights()
        
        for metric in cb.performance.keys():
            if cb.performance[metric] is not None:
                history[metric] += cb.performance[metric]
    
    
    return measure, history, optimal_weights
        
def get_optimal_op_set(search_results, convergence_measure, direction):
    """Select the optimal operator set based on search results
    
    Args:
        search_results (list): List of search results of all operator sets
        convergence_measure (str): Name of the metric to decide optimality
        direction (str): String indicates relative quality of convergence_measure ('higher'/'lower')
    
    Returns:
        performance (dict): Performances of all metrics of optimal operator set
        weights (dict): Corresponding weights of the optimal operator set
        op_set (tuple): Optimal operator set
        history (dict): Full history when evaluting optimal operator set
        
        
    """
    measure = np.NINF if direction == 'higher' else np.inf
    sign = 1.0 if direction == 'higher' else -1.0
    performance = None
    weights = None
    op_set_idx = None
    history = None
    
    for r in search_results:
        if sign*(r[0][convergence_measure]-measure) > 0:
            measure = r[0][convergence_measure]
            performance = r[0]
            weights = r[1]
            op_set_idx = r[2]
            history = r[3]
    
    return performance, weights, op_set_idx, history


def finetune(model_data, 
             params,
             train_func, 
             train_data,
             val_func=None, 
             val_data=None,
             test_func=None,
             test_data=None):
    """Construct and finetune the given network
    
    Args:
        model_data (dict): Data of the model
        params (dict): hyper-parameters to finetune the model
        train_func (function): Function to return train generator and #mini-batches
        train_data: Input to train_func
        val_func (function): Function to return val generator and #mini-batches
        val_data: Input to val_func
        test_func (function): Function to return test generator and #mini-batches
        test_data: Input to test_func
        
    Returns:
        history (dict): Full history during finetuning
        performance (dict): Performance of all metrics at the best model setting
        model_data (dict): Updated model data if performance improves after finetuning
        
    """
    
    
    original_convergence_measure = params['convergence_measure']        
    if val_func:
        params['convergence_measure'] = 'val_' + params['convergence_measure'] 
    else:
        params['convergence_measure'] = 'train_' + params['convergence_measure']
    
    if model_data is None:
        raise Exception('No model data exists for finetuning \n Try load(filename) to load a trained model \n or train a new model with fit() or progressive_learn()')
        
    if params['input_dim'] != model_data['topology'][0]:
        raise Exception('given data has input dimension ´´%d´´ but existing model has input dimension ´´%d´´' %(params['input_dim'], model_data['topology'][0]))
    
    if params['output_dim'] != model_data['topology'][-1][1]:
        raise Exception('given data has output dimension ´´%d´´ but existing model has output dimension ´´%d´´' %(params['output_dim'], model_data['topology'][-1]))
    
    misc.test_generator(train_func, train_data, params['input_dim'], params['output_dim'])
    if val_func:    misc.test_generator(val_func, val_data, params['input_dim'], params['output_dim'])
    if test_func:   misc.test_generator(test_func, test_data, params['input_dim'], params['output_dim'])
    

    model = network_builder(model_data['topology'],
                                      model_data['op_sets'],
                                      params['input_dropout'],
                                      params['dropout_finetune'],
                                      params['weight_regularizer_finetune'],
                                      params['weight_constraint_finetune'],
                                      model_data['output_activation'],
                                      model_data['use_bias'])
    
    model.compile(params['optimizer'], params['loss'], params['metrics'])
        
    model = misc.set_model_weights(model, model_data['weights'])
        
    train_gen, train_steps = train_func(train_data)
    val_gen, val_steps = val_func(val_data) if val_func else (None, None)
    test_gen, test_steps = test_func(test_data) if test_func else (None, None)
        
    performance = misc.evaluate(model,
                                train_gen,
                                train_steps,
                                val_gen,
                                val_steps,
                                test_gen,
                                test_steps)
    
    if params['special_metrics']:
        train_performance = misc.evaluate_special_metrics(model, params['special_metrics'], train_gen, train_steps)
        val_performance = misc.evaluate_special_metrics(model, params['special_metrics'], val_gen, val_steps) if val_gen else None
        test_performance = misc.evaluate_special_metrics(model, params['special_metrics'], test_gen, test_steps) if test_gen else None
        
        for idx, metric in enumerate(params['special_metrics']):
            performance['train_' + metric.__name__] = train_performance[idx]
            performance['val_' + metric.__name__] = val_performance[idx] if val_performance else None
            performance['test_' + metric.__name__] = test_performance[idx] if test_performance else None
    
    measure, history, weights = network_trainer(model, 
                                                params['direction'], 
                                                params['convergence_measure'], 
                                                params['lr_finetune'], 
                                                params['epoch_finetune'], 
                                                params['optimizer'], 
                                                params['loss'], 
                                                params['metrics'], 
                                                params['special_metrics'],
                                                train_func, 
                                                train_data, 
                                                val_func, 
                                                val_data, 
                                                test_func, 
                                                test_data)
    
    sign = 1.0 if params['direction']=='higher' else -1.0
    
    if sign*(measure - performance[params['convergence_measure']]) > 0:
        model.set_weights(weights)
        
        for layer_name in model_data['weights'].keys():
            model_data['weights'][layer_name] = model.get_layer(layer_name).get_weights()
        
        idx = np.argmax(history[params['convergence_measure']]) if params['direction'] == 'higher' else np.argmin(history[params['convergence_measure']])

        for metric in performance.keys():
            if performance[metric] is not None:
                performance[metric] = history[metric][idx]
        
    params['convergence_measure'] = original_convergence_measure
    
    from keras import backend as K
    K.clear_session()
    del K
    
    return history, performance, model_data
    
def evaluate(model_data, func, data, metrics, special_metrics=None):
    """Construct and evaluate the network with given model configuration and data
    
    Args:
        model_data (dict): Data of the model
        func (callable): Function to generate data generator and #mini-batch
        data: Input to func
        metrics (list): List of metrics to evaluate
        special_metrics (list): List of special metrics to evaluate
        
    Returns:
        performance (dict): Performance of given metrics
        
    """
    
    if model_data is None:
        raise Exception('No model exist for evaluation')
    
    misc.test_generator(func, data, model_data['topology'][0], model_data['topology'][-1][1])
    

    model = network_builder(model_data['topology'], 
                                     model_data['op_sets'], 
                                     None, 
                                     None, 
                                     None, 
                                     None, 
                                     model_data['output_activation'], 
                                     model_data['use_bias'])
    
    model.compile('sgd','mse', metrics)
    model = misc.set_model_weights(model, model_data['weights'])
        
    gen, steps = func(data)
    result = model.evaluate_generator(gen, steps)
    performance = {}
    
    for idx, metric in enumerate(model.metrics_names):
        performance[metric] = result[idx]
        
    if special_metrics:
        p = misc.evaluate_special_metrics(model, special_metrics, gen, steps)
        for idx, metric in enumerate(special_metrics):
            performance[metric.__name__] = p[idx]
            
    from keras import backend as K
    K.clear_session()
    del K
        
    return performance

def predict(model_data, func, data):     
    """Construct the network and generate prediction
    
    Args:
        model_data (dict): Data of the network
        func (callable): Function to produce the generator and #mini-batch
        data: Input to func
        
    Returns:
        Numpy array of predicted output
        
    """
        
    if model_data is None:
        raise Exception('No model exist for evaluation')
    
    misc.test_generator(func, data, model_data['topology'][0])
    

    model = network_builder(model_data['topology'], 
                                     model_data['op_sets'], 
                                     None, 
                                     None, 
                                     None, 
                                     None, 
                                     model_data['output_activation'], 
                                     model_data['use_bias'])
    
    model.compile('sgd','mse', ['mse',])
    model = misc.set_model_weights(model, model_data['weights'])
    
    gen, steps = func(data)
    return model.predict_generator(gen, steps)


def load(filename, model_data_attributes, model_name):
    """Load a pretrained model from disk with the given filename
    
    Args:
        filename (str): Path to model data
        model_data_attributes (list): Attributes required from the file
        model_name (str): Name of the model to check compatibility
        
    Returns:
        model_data (dict): Data of the pretrained model
        
    """
    
    try:
        fid = open(filename,'rb')
        model_data = dill.load(fid)
        fid.close()
    except:
        raise Exception('Cannot open model data from ´´filename´´' % filename)
    
    
    model_data_attributes = model_data.keys()
    for key in model_data_attributes:
        if not key in model_data_attributes:
            raise Exception('given model data lacks attribute ´´%s´´' % key)
    
    
    if model_data['model'] != model_name:
        raise Exception('given model data is an instance of ´´%s´´, not %s model' % (model_data['model'], model_name))
    
    model = None    

    try:
        model = network_builder(model_data['topology'], 
                                          model_data['op_sets'], 
                                          None, 
                                          None, 
                                          None, 
                                          None, 
                                          model_data['output_activation'], 
                                          model_data['use_bias'])
        
        model.compile('sgd','mse',['mse',])
    except:
        raise Exception('Cannot build model from model data')
    
    try:
        model = misc.set_model_weights(model, model_data['weights'])
    except:
        raise Exception('Given model weights is incompatible with given topology')
        
    if model is not None:
        from keras import backend as K
        K.clear_session()
        del K
    
    return model_data



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Runable to perform operator set search procedure

Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""
from __future__ import print_function

import sys, getopt, os

module_path = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

if not module_path in sys.path:
    sys.path.append(module_path)


from GOP.utility import gop_utils, misc


try:
    import cPickle as pickle
except ImportError:
    import pickle

def main(argv):

    try:
      opts, args = getopt.getopt(argv,"N:n:")
    except getopt.GetoptError:
        print('gop_search.py -N <number of machine> -n <machine index>')
        sys.exit(2)
        
    for opt, arg in opts:
        if opt == '-N':
            no_machine = int(arg)
        if opt == '-n':
            machine_no = int(arg)
            
    path = os.environ['gop_search_path']
    
    with open(os.path.join(path, 'params.pickle'), 'rb') as f:
        params = pickle.load(f)
        
    no_op_set = params['no_op_set']
    
    start_idx, stop_idx = misc.get_op_set_index(no_op_set, no_machine, machine_no)
    
    if params['search_computation'][0] == 'cpu':
        with open(os.path.join(path, 'train_states_tmp.pickle'),'rb') as f:
            train_states = pickle.load(f)
            
        
        if train_states['model'] in ['HeMLGOP', 'HoMLGOP', 'HeMLRN', 'HoMLRN']:
            search_routine = gop_utils.IRS
        elif train_states['model'] in ['POPfast', 'POPmemO', 'POPmemH']:
            search_routine = gop_utils.GISfast
        elif train_states['model'] == 'POP':
            search_routine = gop_utils.GIS
        
        
        performance, weights, op_set_idx, history = search_routine(start_idx,
                                                                stop_idx,
                                                                params, 
                                                                train_states)
                    
    if params['search_computation'][0] == 'gpu':
            
        performance, weights, op_set_idx, history = gop_utils.search_gpu_(path,
                                                                     params,
                                                                     start_idx,
                                                                     stop_idx,
                                                                     machine_no)

            
    """ write results """
    
    result = {'performance':performance,
              'weights':weights,
              'op_set_idx':op_set_idx,
              'history': history}
    
    with open(os.path.join(path, 'machine_%d.pickle' % machine_no), 'wb') as fid:
        pickle.dump(result, fid)
    
    
    with open(os.path.join(path, 'machine_result.txt'), 'a') as fid:
        fid.write('%d, %.4f \n' % (machine_no, performance[params['convergence_measure']]))
    
    
    with open(os.path.join(path, 'machine_%d_finish.txt' % machine_no), 'w') as fid:
        fid.write('x')
    
    
if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Miscellaneous functions shared by all GOP-based models


Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""

from __future__ import print_function

import os, sys, dill, inspect, copy, numpy as np
dill.settings['recurse'] = True
try:
    import cPickle as pickle
except ImportError:
    import pickle
    
from keras import backend as K
from keras.callbacks import Callback as Callback
from . import gop_operators

class BestModelLogging(Callback):
    """Custom callback used when training Keras model
    
    This callback allows recording all metrics in keras model
    as well as special metrics, which require full batch evaluation instead of
    mini-batch aggregation
    
    Args:
        convergence_measure (str): Name of metric to monitor best model setting
        direction (str): String to indicate relative quality of convergence measure
        special_metrics (list): List of callable to calculate special metrics
        train_func (function): Function to return train generator and #mini-batches
        train_data: Input to train_func
        val_func (function): Function to return val generator and #mini-batches
        val_data: Input to val_func
        test_func (function): Function to return test generator and #mini-batches
        test_data: Input to test_func
        
    Attributes:
        measure (float): Best performance according to the convergence_measure
        performance (dict): History of all metrics for all epochs
        model_weights (list): Weights of the model to achieve best performance (measure)
        
    """
 
    def __init__(self, 
                 convergence_measure,
                 direction,
                 special_metrics=None,
                 train_func=None,
                 train_data=None,
                 val_func=None,
                 val_data=None,
                 test_func=None,
                 test_data=None):
        

        self.convergence_measure = convergence_measure    
        self.direction = direction
        self.special_metrics = special_metrics
        self.performance = {}
        self.train_gen, self.train_steps = train_func(train_data) if train_func else (None, None)
        self.val_gen, self.val_steps = val_func(val_data) if val_func else (None, None)
        self.test_gen, self.test_steps = test_func(test_data) if test_func else (None, None)
        self.special_metrics_names = [metric.__name__ for metric in special_metrics] if special_metrics else []
        
        
    def on_train_begin(self, logs={}):
        self.model_weights = None
        if self.direction == 'higher':
            self.measure = np.NINF
            self.sign = 1.0
        else:
            self.measure = np.inf
            self.sign = -1.0 
        
        for metric in self.model.metrics_names + self.special_metrics_names:
            self.performance['train_' + metric] = []
            self.performance['val_' + metric] = [] if self.val_gen else None
            self.performance['test_' + metric] = [] if self.test_gen else None
                
        self.output_shape = self.model.layers[-1].output_shape[1:]
        
        return
  
    def on_epoch_end(self, epoch, logs={}):
        # log standard metrics
        for metric in self.model.metrics_names:
            self.performance['train_' + metric].append(logs[metric])
        
            if self.val_gen:
                self.performance['val_' + metric].append(logs['val_' + metric])
            
        if self.test_gen:
            test_performance = self.model.evaluate_generator(self.test_gen, self.test_steps)
            for idx, metric in enumerate(self.model.metrics_names):
                self.performance['test_' + metric].append(test_performance[idx])
        
        
        # log special metrics
        if self.special_metrics:
            train_performance = evaluate_special_metrics(self.model, self.special_metrics, self.train_gen, self.train_steps)
            val_performance = evaluate_special_metrics(self.model, self.special_metrics, self.val_gen, self.val_steps) if self.val_gen else None
            test_performance = evaluate_special_metrics(self.model, self.special_metrics, self.test_gen, self.test_steps) if self.test_gen else None
            
            for idx, metric in enumerate(self.special_metrics_names):
                self.performance['train_' + metric].append(train_performance[idx])
                if self.val_gen:
                    self.performance['val_' + metric].append(val_performance[idx])
                if self.test_gen:
                    self.performance['test_' + metric].append(test_performance[idx])
                 
        if self.sign*(self.performance[self.convergence_measure][-1] - self.measure) > 0:
            self.measure = self.performance[self.convergence_measure][-1]
            self.model_weights = self.model.get_weights()
        
        return


def get_he_init(fan_in, fan_out):
    """He initializer method
    
    Args:
        fan_in (int): input dimension
        fan_out (int): output dimension
        
    Returns:
        Callable that allows the initialization with given shape
        
    """
    scale=1.0/max(1.0,float(fan_in+fan_out)/2)
    stddev=np.sqrt(scale)

    def f(shape,dtype=None):
        return K.truncated_normal(shape,stddev=stddev,dtype=dtype)
    return f


def initialize_states(parameters, model):
    """Initialize train states of the given model using the given parameters
    
    This function will check if train_states exists in the temporary dir, load
    the previous state if conditions match, create fresh train_states otherwise
    
    Args:
        parameters (dict): Model parameters
        model (str): Name of the model
        
    Returns:
        train_states (dict): Current model configuration
    
    """

    def get_train_states():
        return {"layer_iter":   0,
                "block_iter":   0,
                "op_set_indices":       {},
                "weights":      {},
                'use_bias':     parameters['use_bias'],
                'output_activation': parameters['output_activation'] if 'output_activation' in parameters.keys() else None,
                "topology":     [parameters['input_dim'],],
                "history":      [],
                'measure':      [],
                "is_finished":  False,
                'model':   model}
    
    path = os.path.join(parameters["tmp_dir"], parameters["model_name"])
    
    if os.path.isdir(path):
        file_path = os.path.join(path, 'train_states.pickle')
        
        if os.path.exists(file_path):
            fid = open(file_path, 'r')
            train_states = pickle.load(fid)
            fid.close()
            
            if train_states['model'] != model or \
            train_states['topology'][0] != parameters['input_dim'] or \
            len(train_states['topology']) == 1 or \
            train_states['topology'][-1][1] != parameters['output_dim'] or \
            train_states['use_bias'] != parameters['use_bias'] or \
            train_states['output_activation'] != parameters['output_activation']:
                train_states = get_train_states()
        else:
            train_states = get_train_states()
    
    else:
        raise Exception('directory "%s" does not exist!' % path)
    
    return train_states

def get_batchjob_parameters_keys():
    compulsory = ['name','mem', 'core', 'partition', 'time', 'no_machine', 'python_cmd']
    noncompulsory = ['configuration', 'constraint']
    
    return compulsory, noncompulsory

def setup_batchjob_file(params, cmd, path):
    """Setup content of batch job file on SLURM cluster using the given settings
    
    Args:
        params (dict): Batch job file parameters
        cmd (str): Command to run python on the cluster
        path (str): Temporary directory that holds output, error files
        
    Returns:
        contents (list): Each element is one row in the batch job file
        
    """
    
    contents = []
    contents.append('#!/bin/bash \n')
    contents.append('#SBATCH -J %s \n' % params['name'])
    contents.append('#SBATCH -o %s' % os.path.join(path, params['name']) + '_%a.o\n')
    contents.append('#SBATCH -e %s' % os.path.join(path,params['name']) + '_%a.e\n')
    contents.append('#SBATCH --mem=%dG \n' % params['mem'])
    contents.append('#SBATCH -c %d \n' % params['core'])
    contents.append('#SBATCH -p %s \n' % params['partition'])
    contents.append('#SBATCH -t %s \n' % params['time'])
    contents.append('#SBATCH --array=0-%s \n' % str(params['no_machine']-1))
    if 'constraint' in params.keys():
        contents.append('#SBATCH --constraint=%s \n' % params['constraint'])
    
    if 'configuration' in params.keys():
        contents.append(params['configuration'] + '\n')
        
    contents.append('\n')
    contents.append(cmd)
    
    return contents

def get_batchjob_parameters_description():
    """Produce description of the batch job parameters required
    
    """
    bj_param_description =  '"name" (str): name of the job submitted to SLURM \n' +\
                                    '"mem" (int): the memory size of each node (in GB) \n' +\
                                    '"core" (int): number of cores of each node\n' +\
                                    '"partition" (str): name of the partition in the cluster \n' +\
                                    '"time" (str): the maximum time to run a node, e.g., "7-00:00:00" is 7 days \n' +\
                                    '"no_machine" (int): number of machines (node) to use \n' +\
                                    '"configuration" (str): the configuration settings, e.g. module loading etc written in string \n' +\
                                    '"python_cmd" (str): command to run python in string, e.g., "srun python" \n' +\
                                    '"constraint" (str): constraint parameter of the cluster, e.g. architecture of cpu or gpu \n'
                                    
    return bj_param_description

def pickle_generator(filename, train_func, train_data, val_func, val_data, test_func, test_data):
    """Pickle the functions that produce generator and steps and dump to disk 
    
    Args:
        filename (str): Filename on disk to dump the function data
        train_func (function): Function to return train generator and #mini-batches
        train_data: Input to train_func
        val_func (function): Function to return val generator and #mini-batches
        val_data: Input to val_func
        test_func (function): Function to return test generator and #mini-batches
        test_data: Input to test_func
        
    Return:
        
    """
    if not os.path.exists(filename):
        data = {'train_func': dill.dumps(train_func, recurse=True),
        'train_data' : train_data,
        'val_func': dill.dumps(val_func, recurse=True) if val_func else None,
        'val_data' : val_data,
        'test_func': dill.dumps(test_func, recurse=True) if test_func else None,
        'test_data' : test_data,
        'train_func_path' : os.path.dirname(inspect.getabsfile(train_func)),
        'val_func_path': os.path.dirname(inspect.getabsfile(val_func)) if val_func else None,
        'test_func_path' : os.path.dirname(inspect.getabsfile(test_func)) if test_func else None}
            
        with open(filename,'wb') as fid:
            pickle.dump(data, fid)
            
    return 

def unpickle_generator(filename):
    """Load the functions to produce data generators from disk
    
    Args:
        filename (str): Path to the pickled functions
        
    Returns:
        train_func (function): Function to return train generator and #mini-batches
        train_data: Input to train_func
        val_func (function): Function to return val generator and #mini-batches
        val_data: Input to val_func
        test_func (function): Function to return test generator and #mini-batches
        test_data: Input to test_func
        
    """

    with open(filename,'rb') as f:
        data = pickle.load(f)
        
    train_gen_path = data['train_func_path']
    val_gen_path = data['val_func_path']
    test_gen_path = data['test_func_path']
    
    if not train_gen_path in sys.path: 
        sys.path.append(train_gen_path)
    
    if val_gen_path and not val_gen_path in sys.path: 
        sys.path.append(val_gen_path)
    
    if test_gen_path and not test_gen_path in sys.path: 
        sys.path.append(test_gen_path)
    
    train_data = data['train_data']
    val_data = data['val_data']
    test_data = data['test_data']
        
    train_func = dill.loads(data['train_func']) if data['train_func'] else None
    val_func = dill.loads(data['val_func']) if data['val_func'] else None
    test_func = dill.loads(data['test_func']) if data['test_func'] else None


    return train_func, train_data, val_func, val_data, test_func, test_data

def get_op_set_index(no_op_set, no_machine, machine_no):
    """Calculate the start and end operator set index that are given to a machine
    
    Args:
        no_op_set: Number of total operator sets
        no_machine: Number of machines used
        machine_no: Machine index
        
    Returns:
        start_idx (int): Start index in the list of operator sets
        stop_idx (int): Stop index in the list of operator sets 
    
    """
    
    assert no_machine <= no_op_set, 'Number of machines (%d) larger than number of operator sets (%d)' %(no_machine, no_op_set)
    op_set_per_machine = int(np.ceil(no_op_set / float(no_machine)))
    start_idx = machine_no*op_set_per_machine
    stop_idx = min(no_op_set, (machine_no+1)*op_set_per_machine)
    
    return start_idx, stop_idx
    
    
def evaluate(model, train_gen, train_steps, val_gen, val_steps, test_gen, test_steps):
    """Evaluate the model using given data generators
    
    Args:
        model: Keras model instance
        train_gen (generator): Train generator
        train_steps (int): Number of mini-batches
        val_gen (generator): Validation generator
        val_steps (int): Number of mini-batches
        test_gen (generator): Test generator
        test_steps (int): Number of mini-batches
        
    Returns:
        performance (dict): Performances according to the metrics inside the Keras model
        
    """
    
    train_performance = model.evaluate_generator(train_gen, train_steps)
    val_performance = model.evaluate_generator(val_gen, val_steps) if val_gen else None
    test_performance = model.evaluate_generator(test_gen, test_steps) if test_gen else None
    
    performance = {}
    for index, metric in enumerate(model.metrics_names):
        performance['train_' + metric] = train_performance[index]
        performance['val_' + metric] = val_performance[index] if val_performance else None
        performance['test_' + metric] = test_performance[index] if test_performance else None
    
    return performance

def evaluate_special_metrics(model, special_metrics, gen, steps):
    """Evaluate the model using given data generators with special metrics
    
    Args:
        model: Keras model instance
        special_metrics (list): List of callable of special metrics
        gen (generator): Train generator
        steps (int): Number of mini-batches
        
    Returns:
        performance (dict): Performances according to the given special metrics
        
    """
    
    y_true = np.zeros([1,] + list(model.layers[-1].output_shape[1:]), dtype=np.float32)
    y_pred = np.zeros([1,] + list(model.layers[-1].output_shape[1:]), dtype=np.float32)
    performance = []
    
    for _ in range(steps):
        x, y = next(gen)
        y_true = np.concatenate((y_true, y), axis=0)
        y_pred = np.concatenate((y_pred, model.predict(x)), axis=0)
    
    y_true, y_pred = y_true[1:], y_pred[1:]
    
    for metric in special_metrics:
        performance.append(metric(y_true, y_pred))
        
    return performance


def check_convergence(new_measure, old_measure, direction, threshold):
    """Check if the performance meets the given threshold
    
    Args:
        new_measure (float): New performance
        old_measure (float): Old performance
        direction (str): String to indicate how to compare two measures
        threshold (float): The given threshold
        
    Returns:
        True if the new measure satisfies threshold, False otherwise
        
    """
    
    sign = 1.0 if direction == 'higher' else -1.0
    
    if sign*(new_measure - old_measure) / old_measure < threshold:
        return True
    else:
        return False
    
def set_model_weights(model, weights):
    """Set the given weights to keras model
    
    Args:
        model : Keras model instance
        weights (dict): Dictionary of weights
    
    Return:
        Keras model instance with weights set
        
    """
    for key in weights.keys():    
        model.get_layer(key).set_weights(weights[key])
        
    return model

    
def test_generator(func, data, input_dim=None, output_dim=None):
    """Test if the given function to produce data generator works
    
    Produce the generators and try to produce data for 10 mini batches
    Raise exception if dimensions mismatch or generator not working
    
    Args:
        func (callable): Function to produce generator
        data: Input to func
        input_dim (int): expected dimension of input
        output_dim (int): optional expected dimension of output
        
    Return:
        
        
    """
    if isinstance(input_dim,int):
        input_dim = (input_dim,)
    if isinstance(output_dim, int):
        output_dim = (output_dim,)
        
    gen, steps = func(data)
    steps = min(steps, 10)
    for _ in range(steps):
        x = next(gen)
        if len(x) == 0:
            raise Exception('output produced by data generator is empty!')
        elif len(x) == 1:
            assert x[0].shape[1:] == input_dim, \
            'input dimension of generated data does not match input dimension of model'
        elif len(x) == 2:
            assert x[0].shape[1:] == input_dim and x[1].shape[1:] == output_dim, \
            'output dimension of generated data does not match output dimension of model'

        
    return

def pickle_custom_metrics(metrics, filename):
    """Pickle the metrics if there is callable in the list of metrics
    
    Args:
        metrics (list): List of metrics
        filename (str): Path to dump the pickled file
        
    Return:
        
    """

    metric_callable = False
    for metric in metrics:
        if callable(metric):
            metric_callable = True
            break
        
    if metric_callable and not os.path.exists(filename):
        metric_names = []
        function_string = []
        function_path = []
        count = 0
        for metric in metrics:
            if isinstance(metric, str):
                metric_names.append(metric)
            elif callable(metric):
                metric_names.append(count)
                function_string.append(dill.dumps(metric, recurse=True))
                
                if not os.path.dirname(inspect.getabsfile(metric)) in function_path:
                    function_path.append(os.path.dirname(inspect.getabsfile(metric)))
                    
                count += 1
                
        metric_data = {'names':metric_names,
                       'function_string': function_string,
                       'function_path': function_path}
        
        with open(filename, 'wb') as fid:
            pickle.dump(metric_data, fid)

    return metric_callable

def unpickle_custom_metrics(filename):
    """Load the pickled metrics 
    
    Args:
        filename (str): Path to the pickled file
        
    Returns:
        metrics (list): List of metrics
        
    """
    metrics = []
    with open(filename, 'rb') as fid:
        metric_data = pickle.load(fid)
        
    names = metric_data['names']
    function_string = metric_data['function_string']
    function_path = metric_data['function_path']
    
    for path in function_path:
        if not path in sys.path:
            sys.path.append(path)
    
    for metric in names:
        if isinstance(metric, str):
            metrics.append(metric)
        elif isinstance(metric, int):
            metrics.append(dill.loads(function_string[metric]))
    
    return metrics

def pickle_custom_loss(loss, filename):
    """Pickle the loss if it is a callable
    
    Args:
        loss : loss function or a string of the function name
        filename (str): Path to dump the pickled file
        
    Return:
        True if loss is callable, False otherwise
        
    """
    
    if callable(loss) and not os.path.exists(filename):
        custom_loss = {}
        custom_loss['function_string'] = dill.dumps(loss, recurse=True)
        custom_loss['path'] = os.path.dirname(inspect.getabsfile(loss))
        
        with open(filename, 'wb') as fid:
            pickle.dump(custom_loss, fid)
        
    if callable(loss):
        return True
    else:
        return False


def unpickle_custom_loss(filename):
    """Load the pickled loss function
    
    Args:
        filename (str): Path to the pickled file
        
    Returns:
        loss (callable): the loss function
        
    """
    
    with open(filename, 'rb') as fid:
        custom_loss = pickle.load(fid)
        
    if not custom_loss['path'] in sys.path:
        sys.path.append(custom_loss['path'])
    
    return dill.loads(custom_loss['function_string'])
    

def pickle_special_metrics(metrics, filename):
    """Pickle special metrics and dump to filename
    
    Args:
        metrics (list): List of special metrics
        filename (str): Path to the pickled file
        
    Return:
        True if special metric is not None, False otherwise
        
    """
    
    if metrics is not None and not os.path.exists(filename):
        data = {'function_string':[],
                'path':[]}
        
        for metric in metrics:
            data['function_string'].append(dill.dumps(metric, recurse=True))
            if not os.path.dirname(inspect.getabsfile(metric)) in data['path']:
                data['path'].append(os.path.dirname(inspect.getabsfile(metric)))
        
        with open(filename, 'wb') as fid:
            pickle.dump(data, fid)
        
    if metrics is not None:
        return True
    else:
        return False

def unpickle_special_metrics(filename):
    """Load the pickled special metrics
    
    Args:
        filename (str): Path to the pickled file
        
    Returns:
        special_metrics (list): List of callable
        
    """
    
    metrics = []
    with open(filename, 'rb') as fid:
        data = pickle.load(fid)
        for path in data['path']:
            if not path in sys.path:
                sys.path.append(path)
                
        for function_str in data['function_string']:
            metrics.append(dill.loads(function_str))
    
    return metrics

def pickle_custom_operators(nodal_set, pool_set, activation_set, filename):
    """Pickle and dump custom operators if exists
    
    Args:
        nodal_set (list): List of nodal operators
        pool_set (list): List of pooling operators
        activation_set (list): List of activation operators
        filename (str): Path to the pickled file on disk
        
    Returns:
        True if exists custom operators else False
        
    """
    
    if os.path.exists(filename):
        return True
    
    has_custom_operators = False
    for operator in nodal_set + pool_set + activation_set:
        if callable(operator):
            has_custom_operators = True
            break
    
    if has_custom_operators:
        count = 0
        pickled_func = []
        nodal_set_ = []
        pool_set_ = []
        activation_set_ = []
        
        for nodal in nodal_set:
            if callable(nodal):
                nodal_set_.append(count)
                count += 1
                pickled_func.append(dill.dumps(nodal, recurse=True))
            else:
                nodal_set_.append(nodal)
                
        for pool in pool_set:
            if callable(pool):
                pool_set_.append(count)
                count += 1
                pickled_func.append(dill.dumps(pool, recurse=True))
            else:
                pool_set_.append(pool)
                
        for activation in activation_set:
            if callable(activation):
                activation_set_.append(count)
                count += 1
                pickled_func.append(dill.dumps(activation, recurse=True))
            else:
                activation_set_.append(activation)
        
        operators = {'nodal_set': nodal_set_,
                     'pool_set': pool_set_,
                     'activation_set': activation_set_,
                     'pickled_func': pickled_func}
        
        with open(filename, 'wb') as fid:
            pickle.dump(operators, fid)
        
    return has_custom_operators


def unpickle_custom_operators(filename):
    """Load the custom operators from pickled file
    
    """
    
    nodal_set = []
    pool_set = []
    activation_set = []
    
    with open(filename, 'rb') as fid:
        operators = pickle.load(fid)
    
    pickled_func = operators['pickled_func']
    
    for nodal in operators['nodal_set']:
        if isinstance(nodal, str):
            nodal_set.append(nodal)
        else:
            nodal_set.append(dill.loads(pickled_func[nodal]))
    
    for pool in operators['pool_set']:
        if isinstance(pool, str):
            pool_set.append(pool)
        else:
            pool_set.append(dill.loads(pickled_func[pool]))
            
    for activation in operators['activation_set']:
        if isinstance(activation, str):
            activation_set.append(activation)
        else:
            activation_set.append(dill.loads(pickled_func[activation]))
            
    return nodal_set, pool_set, activation_set

def get_gpu_str(devices):
    """Produce string of GPUs given the list of int
    
    """
    gpu_str = ''
    for gpu in devices:
        gpu_str += str(gpu) + ','
    return gpu_str[:-1]

def partition_indices(start_idx, stop_idx, no_partition):
    """Partition the range of indices according to given number of partitions
    
    Args:
        start_idx (int): Start index in the range
        stop_idx (int): Stop index in the range
        no_partition (int): Number of partitions
        
    Return:
        start_indices (list): list of start indices
        stop_indices (list): List of stop indices
        
    """
    assert stop_idx > start_idx, 'stop_idx (%d) must be larger than start_idx (%d)' %(stop_idx, start_idx)
    
    start_indices = []
    stop_indices = []
    
    if no_partition > stop_idx - start_idx:
        no_partition = stop_idx - start_idx
    
    step = int(np.ceil(stop_idx-start_idx)/float(no_partition))
    
    for k in range(no_partition):
        start_indices.append(start_idx + k*step)
        stop_indices.append(min(stop_idx, start_idx + (k+1)*step))
        
    return start_indices, stop_indices

def remove_files(filenames):
    """Function to remove the files given in filenames
    
    """
    for f in filenames:
        try:
            os.remove(f)
        except:
            raise Exception('Cannot remove temporary file "%s"' % f)
            
def dump_data(params,
              train_states,
              train_func,
              train_data,
              val_func,
              val_data,
              test_func,
              test_data):
    """Pickle and dump the model parameters, current train states and data functions
    
    Args:
        params (dict): Model parameters
        train_states (dict): Current topology configuration
        train_func (function): Function to return train generator and #mini-batches
        train_data: Input to train_func
        val_func (function): Function to return val generator and #mini-batches
        val_data: Input to val_func
        test_func (function): Function to return test generator and #mini-batches
        test_data: Input to test_func
        
    Returns:
        
        
    """
    
    
    path = os.path.join(params['tmp_dir'], params['model_name'])
    
    params_filename = os.path.join(path, 'params.pickle')
    train_states_filename = os.path.join(path, 'train_states_tmp.pickle')
    data_filename = os.path.join(path, 'data.pickle')
    custom_loss_filename = os.path.join(path, 'custom_loss.pickle')
    custom_metrics_filename = os.path.join(path, 'custom_metrics.pickle')
    special_metrics_filename = os.path.join(path, 'special_metrics.pickle')
    custom_operator_filename = os.path.join(path, 'custom_operators.pickle')
    
    # dump train_states
    with open(train_states_filename, 'wb') as fid:
        pickle.dump(train_states, fid)
            
    # dump custom loss, metrics
    pickle_generator(data_filename, train_func, train_data, val_func, val_data, test_func, test_data)
    has_custom_loss = pickle_custom_loss(params['loss'], custom_loss_filename)
    has_custom_metrics = pickle_custom_metrics(params['metrics'], custom_metrics_filename)
    has_special_metrics = pickle_special_metrics(params['special_metrics'], special_metrics_filename)
    has_custom_operators = pickle_custom_operators(params['nodal_set'], params['pool_set'], params['activation_set'], custom_operator_filename)
    
    if not os.path.exists(params_filename):
        params_ = copy.deepcopy(params)
        if has_custom_loss:
            params_['loss'] = []
        if has_custom_metrics:
            params_['metrics'] = []
        if has_special_metrics:
            params_['special_metrics'] = []
        if has_custom_operators:
            params_['nodal_set'] = []
            params_['pool_set'] = []
            params_['activation_set'] = []
            
        with open(params_filename,'wb') as fid:
            pickle.dump(params_, fid)
    
    return 

def map_operator_from_index(op_set_indices, nodal_set, pool_set, activation_set):
    """Map operator set indices to actual operator set
    
    Args:
        op_set_indices (dict): dictionary of operator set indices of all blocks
        nodal_set (list): List of all nodal operators
        pool_set (list): List of all pooling operators
        activation_set (list): List of all activation operators
        
    Returns:
        op_sets (dict): dictionary of operator sets with same keys
        
    """
    all_op_sets = gop_operators.get_op_set(nodal_set, pool_set, activation_set)
    
    op_sets = {}
    
    for key in op_set_indices.keys():
        op_sets[key] = all_op_sets[op_set_indices[key]]
        
    return op_sets

def check_model_parameters(params, default_params):
    """Perform sanity check on model parameters 
    
    Args:
        params (dict): The given paramters
        default_params (dict): Default parameters of the model
        
    Returns:
        verified params with added default values if missing
        
    """
    keys = params.keys()
    assert os.path.exists(params['tmp_dir']), 'Given temporary directory "%s" does not exist!' % params['tmp_dir']
    assert 'tmp_dir' in keys, 'This model requires a read/writeable temporary directory during computation, please specify in "tmp_dir"'
    assert 'model_name' in keys, 'This model requires a unique name in temporary directory (tmp_dir), please specify in "model_name"'
    assert os.access(params['tmp_dir'], os.W_OK), 'The given temporarary directory "%s" does not have write permission ' % params['tmp_dir']
    
    if not os.path.exists(os.path.join(params['tmp_dir'], params['model_name'])):
        os.mkdir(os.path.join(params['tmp_dir'], params['model_name']))
        
    # verify input_dim, output_dim
    assert 'input_dim' in keys, 'Input dimension should be specified through "input_dim"'
    assert 'output_dim' in keys, 'Output dimension should be specified through "output_dim"'
    
    # setup default parameters if necessary
    
    for key in default_params.keys():
        if not key in keys:
            params[key] = default_params[key]
            
    # verify operator set
    gop_operators.has_valid_operator(params['nodal_set'], params['pool_set'], params['activation_set'])
    params['no_op_set'] = len(params['nodal_set']) * len(params['pool_set']) * len(params['activation_set'])
    
    # verify cluster option
    if params['cluster']:
        description = get_batchjob_parameters_description()
        assert 'batchjob_parameters' in keys, 'When using SLURM cluster, a dictionary of batchjob parameters must be given via the key "batchjob_parameters" \n' +\
                                              '------------------------------------------------------------------------------------------------------------ \n' +\
                                                        'batchjob_parameters should have the following (key, value): \n%s' % description
                                                    
    
        for key in get_batchjob_parameters_keys()[0]:
            if not key in params['batchjob_parameters'].keys():
                raise 'please specify parameter "%s" in "batchjob_parameters"' % key
        
        no_op_set = len(params['nodal_set']) * len(params['pool_set']) * len(params['activation_set'])
        if params['batchjob_parameters']['no_machine'] > no_op_set:
            print('Number of machines used in cluster is larger than number of operator sets, ' + \
                  'setting "no_machine" equal to number of operator sets now')
            params['batchjob_parameters']['no_machine'] = no_op_set
        
    # verify computation devices
    assert isinstance(params['search_computation'], (list, tuple)), '"search_computation" should be given as a tuple or list ' +\
            'with 1st element being "cpu" or "gpu" and 2nd element being number of parallel processes when using "cpu" or a list of GPU numbers when using "gpu"'
            
    if params['search_computation'][0] == 'cpu':
        assert isinstance(params['search_computation'][1], int), 'When "search_computation"="cpu", specify number of parallel processes to used as an integer'
    elif params['search_computation'][0] == 'gpu':
        assert isinstance(params['search_computation'][1], (list, tuple)), 'When "search_computation"="gpu", specify list of gpu number (int) to use'
        
    if params['finetune_computation'][0] == 'gpu':
        assert isinstance(params['finetune_computation'][1], (list, tuple)), 'When "finetune_computation"="gpu", specify list of gpu number (int) to use'
        
    if not 'output_activation' in keys:
        params['output_activation'] = None
    
    assert isinstance(params['metrics'], (str, list, tuple)), 'metrics should be given as a single/list/tuple of (callable or string)'
    if isinstance(params['metrics'], str):
        params['metrics'] = [params['metrics'],]
        
    if params['special_metrics'] is not None:
        assert isinstance(params['special_metrics'], (list, tuple)), 'special_metrics should be given as a list/tuple of callable'
    
    for metric in params['special_metrics']:
        assert callable(metric), 'special_metrics should be a list/tuple of callable'
        
    # correct the metric names if abbreviation is used
    if 'mse' in params['metrics']:
        params['metrics'][params['metrics'].index('mse')] = 'mean_squared_error'
        
    if params['convergence_measure'] == 'mse':
        params['convergence_measure'] = 'mean_squared_error'
        
    if 'accuracy' in params['metrics']:
        params['metrics'][params['metrics'].index('accuracy')] = 'acc'
    
    if params['convergence_measure'] == 'accuracy':
        params['convergence_measure'] == 'acc'

    
    if 'mae' in params['metrics']:
        params['metrics'][params['metrics'].index('mae')] = 'mean_absolute_error'
    
    if params['convergence_measure'] == 'mae':
        params['convergence_measure'] = 'mean_absolute_error'
    
    
    assert isinstance(params['convergence_measure'], str), 'convergence_measure should be a string,' +\
        'in case it is a custom metric or special metric given as a function, convergence_measure should be the function name'
        
    # verify if convergence_measure in metrics
    metrics = []
    for m in params['metrics']:
        if callable(m): 
            metrics.append(m.__name__)
        else:
            metrics.append(m)
    
    if params['special_metrics'] is not None:
        for m in params['special_metrics']:
            metrics.append(m.__name__)
    
    assert params['convergence_measure'] in metrics, 'convergence_measure "%s" should belong to the list of metrics or special_metrics' % params['convergence_measure']
    
    assert isinstance(params['loss'], str) or callable(params['loss']), 'loss should be a string indicate the loss function name or a tensorflow/keras function'
    
    assert params['direction'] in ['higher', 'lower'], 'direction should be ("higher"/"lower") indicating whether higher/lower value of convergence measure is better'
    
    return params


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""

Definition of GOP operators and related utilities


Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""

import tensorflow as tf
global _FLOATX
_FLOATX = str('float64')

def multiplication_(x, w):
    return x*w

def exponential_(x, w):
    y = x*w
    y = tf.clip_by_value(y, -10.0, 10.0)
    y = tf.exp(y) - 1.0    
    return y

def harmonic_(x, w):
    return tf.sin(x*w)

def quadratic_(x, w):
    return tf.square(tf.clip_by_value(x, -10.0, 10.0)*w)

def gaussian_(x, w):
    y = tf.square(x)*w
    y = tf.clip_by_value(y, -10.0, 10.0)
    return w*tf.exp(-y)

def dog_(x, w):
    y = tf.clip_by_value(tf.square(tf.clip_by_value(x, -10.0, 10.0))*w, -10.0, 10.0)
    return w*x*tf.exp(-y)

def sum_(x):
    return tf.reduce_sum(x, axis=-2)

def corr1_(x):
    y = tf.pad(x[:,1:,:], paddings=((0,0), (0,1), (0,0)))
    return tf.reduce_sum(x*y, axis=1)

def corr2_(x):
    y = tf.pad(x[:,1:,:], paddings=((0,0), (0,1), (0,0)))
    z = tf.pad(x[:,2:,:], paddings=((0,0), (0,2), (0,0)))
    return tf.reduce_sum(x*y*z, axis=1)

def max_(x):
    return tf.reduce_max(x, axis=-2)

def sigmoid_(x):
    return tf.sigmoid(x)

def tanh_(x):
    return tf.tanh(x)

def relu_(x):
    return tf.nn.relu(x)

def soft_linear_(x):
    y = tf.clip_by_value(x, -10.0, 10.0)
    return tf.log(1.0 + tf.exp(-y))

def inverse_absolute_(x):
    return x / (1.0 + tf.abs(x))

def exp_linear_(x):
    return tf.nn.elu(x)


    
def get_default_nodal_set():
    """Return the names of supported nodal operators
    
    Returns:
        a tuple of nodal operator names
    """
    output = ['multiplication','exponential','harmonic','quadratic','gaussian','dog']
    #output = ('harmonic','multiplication')
    
    return output

def get_default_pool_set():
    """Return the names of supported pooling operators
    
    Returns:
        a tuple of pooling operator names
    """
    output = ['sum','correlation1','correlation2','maximum']
    #output = ('sum', 'correlation1', 'correlation2','maximum')
    
    return output

def get_default_activation_set():
    """Return the names of supported activation operators
    
    Returns:
        a tuple of activation operator names
    """
    output = ['sigmoid','relu','tanh','soft_linear', 'inverse_absolute', 'exp_linear']
    #output = ('sigmoid','relu', 'tanh', 'inverse_absolute')
    
    return output

def get_op_set(nodal_set, pool_set, activation_set):
    """Create tuple of operator set (in tuple)
    
    """
    op_set = [(nodal, pool, act) for nodal in nodal_set for pool in pool_set for act in activation_set]
    
    return tuple(op_set)

def get_random_op_set():
    """Get a random operator set
    
    """
    import random
    return (random.choice(get_default_nodal_set()), 
            random.choice(get_default_pool_set()), 
            random.choice(get_default_activation_set()))

def get_nodal_operator(instance):
    """Get nodal operator in a callable
    
    """
    if isinstance(instance, str):
        supported_operator = get_default_nodal_set()
        if instance in supported_operator:
            if instance == 'multiplication':
                return multiplication_
            elif instance == 'exponential':
                return exponential_
            elif instance == 'harmonic':
                return harmonic_
            elif instance == 'quadratic':
                return quadratic_
            elif instance == 'gaussian':
                return gaussian_
            elif instance == 'dog':
                return dog_
            else:
                raise "Nodal names mismatch, given ´´%s´´, but only supported ´´%s´´" %(instance, supported_operator)
        else:
            raise "Nodal operator ´´%s´´ not supported!"%instance
    
    else:
        assert callable(instance), "The given nodal operator is neither in default list nor callable"
        return instance
    
def get_pool_operator(instance):
    """Get pooling operator in callable format
    
    """
    if isinstance(instance, str):
        supported_operator = get_default_pool_set()
        if instance in supported_operator:
            if instance == 'sum':
                return sum_
            elif instance == 'correlation1':
                return corr1_
            elif instance == 'correlation2':
                return corr2_
            elif instance == 'maximum':
                return max_
            else:
                raise "Pool names mismatch, given ´´%s´´, but only supported ´´%s´´" %(instance, supported_operator)
        else:
            raise "Pool operator ´´%s´´ not supported!"%instance
    
    else:
        assert callable(instance), "The given pooling operator is neither in default list nor callable"
        return instance
    
def get_activation_operator(instance):
    """Get activation operator in callable format
    
    """
    if isinstance(instance, str):
        supported_operator = get_default_activation_set()
        if instance in supported_operator:
            if instance == 'sigmoid':
                return sigmoid_
            elif instance == 'tanh':
                return tanh_
            elif instance == 'relu':
                return relu_
            elif instance == 'soft_linear':
                return soft_linear_
            elif instance == 'inverse_absolute':
                return inverse_absolute_
            elif instance == 'exp_linear':
                return exp_linear_
            else:
                raise "Activation names mismatch, given ´´%s´´, but only supported ´´%s´´" %(instance, supported_operator)
        else:
            raise "Activation operator ´´%s´´ not supported!"%instance
    
    else:
        assert callable(instance), "The given activation operator is neither in default list nor callable"
        return instance


def has_valid_operator(nodal_set, pool_set, activation_set):
    """Check if the names of operator are valid
    
    Args:
        nodal_set (tuple): names of nodal operator
        pool_set (tuple): names of pooling operator
        activation_set (tuple): names of activation operator
        
    Returns:
        
    
    """
    
    supported_nodal = get_default_nodal_set()
    supported_pool = get_default_pool_set()
    supported_activation = get_default_activation_set()
    
    for nodal in nodal_set:
        if isinstance(nodal, str):
            if not nodal in supported_nodal:
                raise 'Nodal operator ´´%s´´ not supported' % nodal
        else:
            assert callable(nodal), "The given nodal operator is neither in default list nor callable"
            

    for pool in pool_set:
        if isinstance(pool, str):
            if not pool in supported_pool:
                raise "Pooling operator ´´%s´´ not supported" % pool
        else:
            assert callable(pool), "The given pooling operator is neither in default list nor callable"

    for activation in activation_set:
        if isinstance(activation, str):
            if not activation in supported_activation:
                raise "Activation operator ´´%s´´ not supported" % pool
        else:
            assert callable(pool), "The given activation operator is neither in default list nor callable"
        
    return


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""

from . import gop_utils
from . import gop_operators
from . import misc



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Heterogeneous Multilayer Generalized Operational Perceptron
https://arxiv.org/abs/1804.05093


Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""
from __future__ import print_function

from ..utility import misc, gop_utils, gop_operators
from ._model import _Model, CUDA_FLAG
import shutil, os, copy
try:
    import cPickle as pickle
except ImportError:
    import pickle
    
    
    
    
class HeMLGOP(_Model):
    """Heterogeneous Multilayer Generalized Operational Perceptron (HeMLGOP) model
    
    This class implements the HeMLGOP algorithm to learn a heterogeneous multilayer 
    network of Generalized Operational Perceptron (GOP) in a progressive manner by
    incrementing the neuron blocks in a layer and incrementing the layers.
    
    Note:
        The model basically uses a python data generator mechanism to feed the data
        Thus, the data should be prepared with the following format:
            train_func: a function that returns (data_generator, number of steps)
                        the user specifies how data is loaded and preprocess by giving
                        the definition of the data generator and the number of steps (number of mini-batches)
            train_data: the input to 'train_func'
                        this can be filepath to the data on disk
            
        When the model generates the data, the generator and #steps are retrieved by
        calling:
            gen, steps = train_func(train_data)
            
        And next(gen) should produce a mini-batch of (x,y) in case of fit()/evaluate() 
        or only x in case of predict()
        
        See documentation page for example how to define such train_func and train_data
        
        
    Examples:
        >>> from GOP import models
        >>> model = models.HeMLGOP() # create a model instance
        >>> params = model.get_default_parameters() # get default parameters
        
        # fit the model with data using train_func, train_data (see above in Note)
        >>> performance, progressive_history, finetune_history = model.fit(params, train_func, train_data)
        
        # evaluate model with test data (test_func, test_data) using a list of metrics
        # and special_metrics (those require full batch evaluation such as F1, Precision, Recall...)
        # using either CPU or GPU as computation environment,
        # e.g. computation=('cpu',) -> using cpu
        # e.g. computation=('gpu', [0,1]) -> using gpu with GPU0, GPU1
        
        >>> model.evaluate(test_func, test_data, metrics, special_metrics, computation) 

        # generate prediction with (test_func, test_data)
        >>> model.predict(test_func, test_data, computation)
        
        # save model to 'hemlgop.pickle'
        >>> model.save('hemlgop.pickle')
        
        # load model from 'hemlgop.pickle' and finetune with another data and
        # (possibly different) parameter settings
        >>> model.load('hemlgop.pickle')
        >>> history, performance = model.finetune(params, another_train_func, another_train_data)
        
        See documentation page for detail usage and explanation
        
        
    """
    
    def __init__(self,):
        self.model_data = None 
        self.model_data_attributes = ['model', 
                                      'topology', 
                                      'op_sets', 
                                      'weights', 
                                      'output_activation', 
                                      'use_bias']
        
        self.model_name = 'HeMLGOP'
        
        return
    
    
    def get_default_parameters(self,):
    
        params = {}
        
        params['least_square_regularizer'] = 0.1
        params['block_threshold'] = 1e-4
        params['layer_threshold'] = 1e-4
        params['metrics'] = ['mse',]
        params['convergence_measure'] = 'mse'
        params['direction'] = 'lower'
        params['output_activation'] = None
        params['loss'] = 'mse'
        params['max_block'] = 5
        params['block_size'] = 20
        params['max_layer'] = 4
        params['weight_regularizer'] = None
        params['weight_regularizer_finetune'] = None
        params['weight_constraint'] = 2.0
        params['weight_constraint_finetune'] = 2.0
        params['lr_train'] = (0.01,0.001,0.0001)
        params['epoch_train'] = (2,2,2)
        params['lr_finetune'] = (0.0005,)
        params['epoch_finetune'] = (2,)
        params['input_dropout'] = None
        params['dropout'] = 0.2
        params['dropout_finetune'] = 0.2
        params['optimizer'] = 'adam'
        params['nodal_set'] = gop_operators.get_default_nodal_set()
        params['pool_set'] = gop_operators.get_default_pool_set()
        params['activation_set'] = gop_operators.get_default_activation_set()
        params['direct_computation'] = False
        params['cluster'] = False
        params['special_metrics'] = None
        params['search_computation'] = ('cpu', 8)
        params['finetune_computation'] = ('cpu', 8)
        params['use_bias'] = True
        
        return  params
    
    def fit(self, 
            params,
            train_func, 
            train_data, 
            val_func=None, 
            val_data=None, 
            test_func=None, 
            test_data=None,
            verbose=False):
        
        if verbose:
            print('Start progressive learning')
        
        p_history = self.progressive_learn(params,
                                              train_func,
                                              train_data,
                                              val_func,
                                              val_data,
                                              test_func,
                                              test_data,
                                              verbose)
        
        if verbose:
            print('Start finetuning')
        
        f_history, performance = self.finetune(params,
                                               train_func,
                                               train_data,
                                               val_func,
                                               val_data,
                                               test_func,
                                               test_data,
                                               verbose)
        
        return performance, p_history, f_history
        
        
        
    def progressive_learn(self,
                          params,
                          train_func, 
                          train_data,
                          val_func=None, 
                          val_data=None,
                          test_func=None,
                          test_data=None,
                          verbose=False):
    
        
        if verbose:
            print('checking parameters')
        params = self.check_parameters(params)

        original_convergence_measure = params['convergence_measure']        
        if val_func:
            params['convergence_measure'] = 'val_' + params['convergence_measure'] 
        else:
            params['convergence_measure'] = 'train_' + params['convergence_measure']
        
        if verbose:
            print('checking generators')
        
        misc.test_generator(train_func, train_data, params['input_dim'], params['output_dim'])
        if val_func:    misc.test_generator(val_func, val_data, params['input_dim'], params['output_dim'])
        if test_func:   misc.test_generator(test_func, test_data, params['input_dim'], params['output_dim'])
        
        train_states = misc.initialize_states(params, self.model_name)
        
        
        if not train_states['is_finished']:
            for layer_iter in range(train_states['layer_iter'], params['max_layer']):
                block_stop = False
                for block_iter in range(train_states['block_iter'], params['max_block']):
                                            
                    if block_iter == 0:
                        if layer_iter > 0:
                            train_states['topology'].pop()
                            
                        train_states['topology'].append([])
                        train_states['topology'].append(('dense', params['output_dim']))
                        train_states['history'].append([])
                        train_states['measure'].append([])
                        
                    if verbose:
                        print('-------------Layer %d ---------------Block %d ------------------' %(layer_iter, block_iter))
                        print('topology')
                        print(train_states['topology'])
                        print('op_sets')
                        print(train_states['op_set_indices'])
                        print('weights')
                        print(train_states['weights'].keys())
                        print('layer iter in train_states ' + str(train_states['layer_iter']))
                        print('block iter in train_states ' + str(train_states['block_iter']))
                        
                    if verbose:
                        print('##### Iterative Randomized Search #####')
                        
                    if params['cluster']:
                        search_routine = gop_utils.search_cluster
                    else:
                        if params['search_computation'][0] == 'cpu':
                            search_routine = gop_utils.search_cpu
                        else:
                            search_routine = gop_utils.search_gpu
                                  
                    _, block_weights, block_op_set_idx, _ = search_routine(params,
                                                                      train_states,
                                                                      train_func,
                                                                      train_data,
                                                                      val_func,
                                                                      val_data,
                                                                      test_func,
                                                                      test_data)
                        
                    
                    # block_weights[gop, bn, output]
                    suffix = '_' + str(layer_iter) + '_' + str(block_iter)
                    block_names = ['gop' + suffix, 'bn' + suffix, 'output']
                    
                    # add new block configuration
                    train_states['topology'][-2].append(('gop',params['block_size']))
                    train_states['op_set_indices']['gop'+suffix] = block_op_set_idx
                    train_states['weights']['gop'+suffix], train_states['weights']['bn'+suffix], train_states['weights']['output'] = block_weights
                    
                    
                    if verbose:
                        print('##### Block finetuning ######')
                    
                    if CUDA_FLAG in os.environ.keys():
                        cuda_status = os.environ[CUDA_FLAG]
                    else:
                        cuda_status = None
                        
                    if params['finetune_computation'][0] == 'cpu':
                        os.environ[CUDA_FLAG] = ''
                    else:
                        os.environ[CUDA_FLAG] = misc.get_gpu_str(params['finetune_computation'][1])
                        
                    try:
                        new_measure, history, train_states['weights'] = gop_utils.block_update(train_states['topology'], 
                                                                                              train_states['op_set_indices'], 
                                                                                              train_states['weights'], 
                                                                                              params, 
                                                                                              block_names, 
                                                                                              train_func,
                                                                                              train_data,
                                                                                              val_func,
                                                                                              val_data,
                                                                                              test_func,
                                                                                              test_data)
                    except:
                        if cuda_status is None:
                            del os.environ[CUDA_FLAG]
                        else:
                            os.environ[CUDA_FLAG] = cuda_status
                        raise Exception('Fail updating block weights after randomized search')
                    
                    if cuda_status is None:
                        del os.environ[CUDA_FLAG]
                    else:
                        os.environ[CUDA_FLAG] = cuda_status
                    
                    if block_iter > 0:
                        if misc.check_convergence(new_measure, train_states['measure'][layer_iter][block_iter-1], params['direction'], params['block_threshold']):
                            train_states['topology'][-2].pop()
                            del train_states['weights']['gop'+suffix]
                            del train_states['weights']['bn'+suffix]
                            del train_states['op_set_indices']['gop'+suffix]
                            train_states['weights']['output'] = copy.deepcopy(train_states['block_output_weight'])
                            
                            train_states['block_iter'] = 0
                            train_states['layer_iter'] += 1
                            block_stop = True
                            path = os.path.join(params['tmp_dir'], params['model_name'], 'train_states.pickle')
                            with open(path, 'wb') as fid:
                                pickle.dump(train_states, fid)
                            
                            break
                        
                    
                    train_states['block_output_weight'] = copy.deepcopy(train_states['weights']['output'])
                    train_states['measure'][layer_iter].append(new_measure)
                    train_states['history'][layer_iter].append(history)
                    train_states['block_iter'] = 0 if block_iter == params['max_block'] -1 else block_iter + 1
                    
                    path = os.path.join(params['tmp_dir'], params['model_name'], 'train_states.pickle')
                    with open(path, 'wb') as fid:
                        pickle.dump(train_states, fid)
                    
                    
                if layer_iter > 0:
                    if misc.check_convergence(train_states['measure'][layer_iter][-1], train_states['measure'][layer_iter-1][-1], params['direction'], params['layer_threshold']):
                        no_blocks = len(train_states['topology'].pop(-2))
                        suffices = ['_' + str(layer_iter) + '_' + str(idx) for idx in range(no_blocks)]
                        for suffix in suffices:
                            del train_states['weights']['gop' + suffix]
                            del train_states['weights']['bn' + suffix]
                            del train_states['op_set_indices']['gop' + suffix]
                        train_states['weights']['output'] = copy.deepcopy(train_states['layer_output_weight'])   
                        
                        break
                
                if not block_stop:
                    train_states['layer_iter'] += 1
                
                train_states['layer_output_weight'] = copy.deepcopy(train_states['weights']['output'])
                    
                    
            train_states['is_finished'] = True
            path = os.path.join(params['tmp_dir'], params['model_name'], 'train_states.pickle')
            with open(path, 'wb') as fid:
                pickle.dump(train_states, fid)             
            
            if os.path.exists(os.path.join(params['tmp_dir'], params['model_name'])):
                shutil.rmtree(os.path.join(params['tmp_dir'], params['model_name']))
            

        model_data = {'model': self.model_name,
                      'topology': train_states['topology'],
                      'op_sets': misc.map_operator_from_index(train_states['op_set_indices'], params['nodal_set'], params['pool_set'], params['activation_set']),
                      'weights': train_states['weights'],
                      'use_bias': train_states['use_bias'],
                      'output_activation': train_states['output_activation']}
        
        self.model_data = model_data
        params['convergence_measure'] = original_convergence_measure
        
        return train_states['history']
                
        

        


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Author: Dat Tran
Email: dat.tranthanh@tut.fi, viebboy@gmail.com
github: https://github.com/viebboy
"""

import setuptools
from GOP.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()
    
    
setuptools.setup(
    name="pygop",
    version=__version__,
    author="Dat Tran",
    author_email="viebboy@gmail.com",
    description="Python package that implements various algorithms using Generalized Operational Perceptron",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/viebboy/PyGOP",
    license='LICENSE.txt',
    packages=setuptools.find_packages(),
    classifiers=['Operating System :: POSIX',],
    install_requires = ['python_version>= "2.7"' or 'python_version >= "3.4"',
                      'numpy >= 1.13',
                      'dill >= 0.2.6',
                      'joblib >= 0.11',
                      'keras >= 2.2.1'],
    setup_requires = ['numpy >= 1.13',
                      'dill >= 0.2.6',
                      'joblib >= 0.11',
                      'keras >= 2.2.1']
)
